import unittest
import numpy as np
import pandas as pd

from inmoose.utils import rnbinom
from inmoose.pycombat import pycombat_seq


class test_pycombatseq(unittest.TestCase):
    def setUp(self):
        y = np.array(rnbinom(100, size=5, mu=20, seed=42)).reshape((20, 5))
        self.y = np.vstack(([0, 0, 0, 0, 0], [0, 0, 2, 2, 2], y))
        self.batch = np.array([1, 1, 2, 2, 2])

    def test_pycombat_seq(self):
        ref = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 2, 2, 2],
                [21, 14, 14, 26, 14],
                [24, 13, 15, 13, 28],
                [29, 5, 13, 3, 30],
                [30, 20, 16, 33, 26],
                [12, 36, 31, 26, 8],
                [10, 20, 17, 8, 24],
                [18, 31, 34, 24, 18],
                [23, 14, 22, 15, 19],
                [32, 15, 11, 32, 33],
                [27, 15, 15, 30, 17],
                [25, 28, 20, 25, 36],
                [28, 14, 26, 15, 18],
                [12, 47, 36, 6, 42],
                [23, 39, 15, 44, 34],
                [31, 20, 26, 24, 26],
                [29, 17, 21, 26, 18],
                [37, 20, 14, 37, 33],
                [21, 19, 13, 28, 21],
                [23, 14, 16, 14, 25],
                [13, 33, 16, 45, 11],
            ]
        )
        res = pycombat_seq(self.y, self.batch)
        self.assertTrue(np.array_equal(res, ref))

        ref = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 2, 2, 2],
                [19, 14, 14, 29, 13],
                [22, 13, 15, 13, 30],
                [34, 2, 15, 8, 25],
                [24, 18, 17, 41, 24],
                [11, 43, 24, 25, 12],
                [14, 21, 16, 7, 23],
                [31, 39, 29, 19, 10],
                [28, 18, 18, 12, 17],
                [17, 12, 11, 45, 47],
                [24, 14, 16, 33, 17],
                [24, 26, 20, 22, 42],
                [34, 18, 21, 13, 16],
                [6, 67, 33, 10, 35],
                [19, 31, 20, 51, 33],
                [34, 23, 23, 21, 24],
                [30, 18, 19, 26, 16],
                [32, 17, 19, 38, 32],
                [17, 16, 14, 36, 20],
                [23, 14, 15, 13, 25],
                [8, 30, 18, 49, 15],
            ]
        )
        res = pycombat_seq(self.y, self.batch, covar_mod=[1, 1, 1, 2, 2])
        self.assertTrue(np.array_equal(res, ref))

        # test with reference batch
        res = pycombat_seq(self.y, self.batch, ref_batch=1)
        ref_col = self.batch == 1
        non_ref_col = self.batch != 1
        # make sure that reference batch counts have not been adjusted
        self.assertTrue(np.array_equal(res[:, ref_col], self.y[:, ref_col]))
        # make sure that batch effects have still been corrected
        self.assertFalse(np.array_equal(res[:, non_ref_col], self.y[:, non_ref_col]))

        # also test with non-integer batch ids
        res2 = pycombat_seq(self.y, ["a", "a", "b", "b", "b"], ref_batch="a")
        self.assertTrue(np.array_equal(res, res2))

        # test raise error for incorect counts format
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="counts must be a pandas DataFrame or a numpy nd array",
        ):
            pycombat_seq(
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 2, 2, 2],
                    [19, 14, 14, 29, 13],
                    [22, 13, 15, 13, 30],
                    [34, 2, 15, 8, 25],
                    [24, 18, 17, 41, 24],
                    [11, 43, 24, 25, 12],
                    [14, 21, 16, 7, 23],
                    [31, 39, 29, 19, 10],
                    [28, 18, 18, 12, 17],
                    [17, 12, 11, 45, 47],
                    [24, 14, 16, 33, 17],
                    [24, 26, 20, 22, 42],
                    [34, 18, 21, 13, 16],
                    [6, 67, 33, 10, 35],
                    [19, 31, 20, 51, 33],
                    [34, 23, 23, 21, 24],
                    [30, 18, 19, 26, 16],
                    [32, 17, 19, 38, 32],
                    [17, 16, 14, 36, 20],
                    [23, 14, 15, 13, 25],
                    [8, 30, 18, 49, 15],
                ],
                ["a", "b", "b", "b", "b"],
            )

        # test raise error for single sample batch
        with self.assertRaisesRegex(
            ValueError, expected_regex="Batches a contain a single sample"
        ):
            pycombat_seq(self.y, ["a", "b", "b", "b", "b"])

        # test with incomplete group
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="2 values are missing in covariates cov_0. Correct your covariates or use the cov_missing_value parameters",
        ):
            pycombat_seq(self.y, self.batch, covar_mod=["a", np.nan, "a", np.nan, "a"])

        ref = pycombat_seq(self.y, self.batch, covar_mod=["a", "b", "a", "c", "a"])
        with self.assertWarnsRegex(
            UserWarning,
            r"1 missing covariates in covar_mod. Creating a distinct covariate per batch for the missing values. You may want to double check your covariates.",
        ):
            res = pycombat_seq(
                self.y,
                self.batch,
                covar_mod=[1, 2, 1, np.nan, 1],
                na_cov_action="fill",
            )
        self.assertTrue(np.array_equal(res, ref))

        with self.assertWarnsRegex(
            UserWarning,
            r"1 samples with missing covariates in covar_mod. They are removed from the data. You may want to double check your covariates.",
        ):
            res = pycombat_seq(
                self.y,
                self.batch,
                covar_mod=[1, 2, 1, np.nan, 1],
                na_cov_action="remove",
            )

        # with count as dataframe
        # check remove option warning message
        with self.assertWarnsRegex(
            UserWarning,
            r"1 samples with missing covariates in covar_mod. They are removed from the data. You may want to double check your covariates.",
        ):
            res = pycombat_seq(
                pd.DataFrame(
                    self.y,
                    columns=["sample1", "sample2", "sample3", "sample4", "sample5"],
                ),
                self.batch,
                covar_mod=pd.DataFrame(
                    [1, 2, 1, np.nan, 1],
                    columns=["test"],
                    index=["sample1", "sample2", "sample3", "sample4", "sample5"],
                ),
                na_cov_action="remove",
            )

        # check remove option results
        res2 = pycombat_seq(
            pd.DataFrame(
                self.y[:, [0, 1, 2, 4]],
                columns=["sample1", "sample2", "sample3", "sample5"],
            ),
            np.array([1, 1, 2, 2]),
            covar_mod=pd.DataFrame(
                [1, 2, 1, 1],
                columns=["test"],
                index=["sample1", "sample2", "sample3", "sample5"],
            ),
        )
        self.assertTrue(np.array_equal(res, res2))

        # check confounded covariates
        with self.assertRaisesRegex(
            ValueError,
            expected_regex=r"Covariates test are confounded with the batches. Please review your covariates before proceeding with batch effect correction.",
        ):
            pycombat_seq(
                self.y,
                self.batch,
                covar_mod=pd.DataFrame(
                    [2, 2, 1, 1, 1],
                    columns=["test"],
                ),
            )

        with self.assertRaisesRegex(
            ValueError,
            expected_regex=r"Covariates cov_0 are confounded with the batches. Please review your covariates before proceeding with batch effect correction.",
        ):
            pycombat_seq(
                self.y,
                self.batch,
                covar_mod=pd.DataFrame(
                    [2, 2, 1, 1, 1],
                ),
            )

        with self.assertWarnsRegex(
            UserWarning,
            r"1 samples with missing covariates in covar_mod. They are removed from the data. You may want to double check your covariates.",
        ):
            res = pycombat_seq(
                self.y,
                self.batch,
                covar_mod=pd.DataFrame(["a", "b", "a", np.nan, "a"], columns=["test"]),
                na_cov_action="remove",
            )

        ref_y = np.delete(self.y, (3), axis=1)
        ref_batch = np.array([1, 1, 2, 2])
        ref = pycombat_seq(ref_y, ref_batch, covar_mod=[1, 2, 1, 1])
        self.assertTrue(np.array_equal(res, ref))

        # test error/warning message for data type of covariates
        with self.assertRaisesRegex(
            ValueError,
            expected_regex=r"Cannot create new categories for numerical covariates cov_0. Please fix the NA in those covariates manually.",
        ):
            pycombat_seq(
                self.y,
                self.batch,
                covar_mod=[1, 2.9, 1, 1, np.nan],
                na_cov_action="fill",
            )



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/estimateGLMCommonDisp.R' of the Bioconductor edgeR package (version 3.38.4).

import logging
import numpy as np
from .aveLogCPM import aveLogCPM
from .dispCoxReid import dispCoxReid
from .validDGEList import validDGEList


def estimateGLMCommonDisp_DGEList(
    self, design=None, method="CoxReid", subset=10000, verbose=False
):
    """
    Estimate a common negative binomial dispersion parameter for a DGE dataset
    with a general experimental design.

    NB: modifies :code:`self` in place

    Arguments
    ---------
    self : DGEList
        the DGEList containing the matrix of counts, as in :func:`glmFit`
    design : matrix, optional
        design matrix, as in :func:`glmFit`
    method : str
        method for estimating the dispersion. Possible values are "CoxReid",
        "Pearson" or "deviance". Defaults to "CoxReid".
    subset : int
        maximum number of rows of :code:`y` to use in the calculation. Rows
        used are chosen evenly spaced by :code:`AveLogCPM` using
        :func:`systematicSubset`.

    Returns
    -------
    DGEList
        :code:`self` updated with :code:`common_dispersion`, and :code:`AveLogCPM`
        if it was not alread present in input :code:`self`.
    """
    y = validDGEList(self)
    AveLogCPM = y.aveLogCPM(dispersion=0.05)

    disp = estimateGLMCommonDisp(
        y=y.counts,
        design=design,
        offset=y.getOffset(),
        method=method,
        subset=subset,
        AveLogCPM=AveLogCPM,
        verbose=verbose,
        weights=y.weights,
    )

    y.common_dispersion = disp
    y.AveLogCPM = y.aveLogCPM(dispersion=disp)
    return y


def estimateGLMCommonDisp(
    y,
    design=None,
    offset=None,
    method="CoxReid",
    subset=10000,
    AveLogCPM=None,
    verbose=False,
    weights=None,
):
    """
    Estimate a common negative binomial dispersion parameter for a DGE dataset
    with a general experimental design.

    This function calls :func:`dispCoxReid`, :func:`dispPearson` or
    :func:`dispDeviance` depending on the :code:`method` specified. See
    :func:`dispCoxReid` for details of the three methods and a discussion of
    their relative performance.

    See also
    --------
    dispCoxReid
    estimateGLMTrendedDisp : for trended dipsersions
    estimateGLMTagwiseDisp : for genewise dispersions in the context of a GLM
    estimateCommonDisp : for the common dispersion
    estimateTagwiseDisp : for genewise dispersion in the context of a multiple
        group experiment (one-way layout)

    Arguments
    ---------
    y : matrix
        matrix of counts, as in :func:`glmFit`
    design : matrix, optional
        design matrix, as in :func:`glmFit`
    offset : array_like, optional
        vector or matrix of offsets for the log-linear models, as in
        :func:`glmFit`
    method : str
        method for estimating the dispersion. Possible values are "CoxReid",
        "Pearson" or "deviance". Defaults to "CoxReid".
    subset : int
        maximum number of rows of :code:`y` to use in the calculation. Rows
        used are chosen evenly spaced by :code:`AveLogCPM` using
        :func:`systematicSubset`.
    AveLogCPM : array_like
        vector of log2 average counts per million for each gene
    weights : matrix, optional
        observation weights

    Returns
    -------
    float
        estimated common dispersion
    """

    # Check design
    if design is None:
        design = np.ones((y.shape[1], 1))
    if design.shape[1] >= y.shape[1]:
        logging.warnings.warn(
            "No residual degree of freedom: setting dispersion to None"
        )
        return None

    # Check method
    if method != "CoxReid" and weights is not None:
        logging.warnings.warn("weights only supported by CoxReid method")

    # Check offset
    if offset is None:
        offset = np.log(y.sum(axis=0))

    # Check AveLogCPM
    if AveLogCPM is None:
        AveLogCPM = aveLogCPM(y, offset, weights)

    # Call lower-level function
    if method == "CoxReid":
        disp = dispCoxReid(
            y,
            design=design,
            offset=offset,
            subset=subset,
            AveLogCPM=AveLogCPM,
            weights=weights,
        )
    elif method == "Pearson":
        disp = dispPearson(
            y, design=design, offset=offset, subset=subset, AveLogCPM=AveLogCPM
        )
    elif method == "deviance":
        disp = dispDeviance(
            y, design=design, offset=offset, subset=subset, AveLogCPM=AveLogCPM
        )
    else:
        raise ValueError(f"invalid method for dispersion evaluation: {method}")

    logging.debug(f"Disp = {round(disp, 5)}, BCV = {round(np.sqrt(disp), 4)}")

    return disp



from .addPriorCount import addPriorCount
from .adjustedProfileLik import adjustedProfileLik
from .aveLogCPM import aveLogCPM
from .DGEGLM import DGEGLM
from .DGEList import DGEList
from .dispCoxReid import dispCoxReid
from .dispCoxReidInterpolateTagwise import dispCoxReidInterpolateTagwise
from .estimateGLMCommonDisp import estimateGLMCommonDisp
from .estimateGLMTagwiseDisp import estimateGLMTagwiseDisp
from .glmFit import glmFit
from .maximizeInterpolant import maximizeInterpolant
from .mglmLevenberg import mglmLevenberg
from .mglmOneGroup import mglmOneGroup
from .mglmOneWay import mglmOneWay, designAsFactor
from .movingAverageByCol import movingAverageByCol
from .nbinomDeviance import nbinomDeviance
from .predFC import predFC
from .splitIntoGroups import splitIntoGroups
from .stats import pnbinom, qnbinom
from .systematicSubset import systematicSubset
from .validDGEList import validDGEList
from .edgepy_cpp import *



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2021 Ž. Sajovic
# Copyright (C) 2023 S. Weill and G. Appé

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'consensusClustering.py' of the repository
# github.com/ZigaSajovic/Consensus_Clustering (as of July 30, 2021).

import logging
import numpy as np
from itertools import combinations
import bisect
import seaborn as sn
import matplotlib.pyplot as plt
import pandas as pd


class consensusClustering:
    """
    Implementation of Consensus clustering, following the paper https://link.springer.com/content/pdf/10.1023%2FA%3A1023949509487.pdf

    Arguments
    ---------
    cluster: sklearn clustering class
        Clustering algorithm to use for consensus clustering
        NOTE: the class is to be instantiated with parameter `n_clusters`, and possess a `fit_predict` method, which is invoked on data.
    mink: int
        smallest number of clusters to try, default = 2
    maxk: int
        biggest number of clusters to try, default = 10
    nb_resampling_iteration: int
        number of resamplings for each cluster number, default = 50
    resample_proportion: float
        percentage to sample. Number between 0 and 1, default = 0.5
    n_bins: int
        Number of bins used to compute histogram in compute_area_under_curve, default = 10
    consensus_matrices: ndarray[float]
        consensus matrices for each k
        NOTE: every consensus matrix is retained, like specified in the paper
    Ak: array[float]
        area under CDF for each number of clusters (see paper: section 3.3.1. Consensus distribution.)
    deltaK: array[float]
        changes in areas under CDF (see paper: section 3.3.1. Consensus distribution.)
    bestK: int
        number of clusters that was found to be best
    """

    def __init__(
        self,
        cluster,
        mink=2,
        maxk=10,
        nb_resampling_iteration=50,
        resample_proportion=0.5,
        n_bins=10,
    ):
        assert 0 <= resample_proportion <= 1, "proportion has to be between 0 and 1"
        self.cluster_ = cluster
        self.resample_proportion = resample_proportion
        self.min_k = mink
        self.max_k = maxk
        self.nb_iteration = nb_resampling_iteration
        self.nbins = n_bins
        self.consensus_matrices = None
        self.Ak = None
        self.deltaK = None
        self.bestK = None

    def _internal_resample(self, data, proportion, rand_state):
        """
        Sampling data based on the proportion of samples

        Arguments
        ---------
        data: ndarray
            numpy matrix to run the cluster on samples*features
        proportion: float
            percentage to sample
        rand_state: numpy.random.Generator
            Numpy random generator instance

        Returns
        -------
        resampled_indices
            sample's indices to use for the iteration
        """

        resampled_indices = rand_state.choice(
            range(data.shape[0]), size=int(data.shape[0] * proportion), replace=False
        )
        return resampled_indices

    def compute_consensus_clustering(self, data, random_state, verbose=False):
        """
        Fits a consensus matrix for each number of clusters

        Arguments
        ---------
        data: ndarray
            numpy matrix to run the cluster on samples*features
        random_state: int
            seed to use to generate a numpy generator instance. Default is None
        verbose: bool (default=False)
            If True, print the number of clusters for which the consensus matrix is computed
        """
        rand_state = np.random.default_rng(random_state)
        consensus_mats = np.zeros(
            (self.max_k - self.min_k + 1, data.shape[0], data.shape[0])
        )

        for k in range(self.min_k, self.max_k + 1):  # for each number of clusters
            if verbose:
                logging.info(f"Computing consensus matrix for {k} clusters")
            indicator_matrix = np.zeros((data.shape[0],) * 2)
            connectivity_matrix = np.zeros((data.shape[0],) * 2)

            for _ in range(self.nb_iteration):  # resample H times
                resampled_indices = self._internal_resample(
                    data, self.resample_proportion, rand_state
                )

                Mh = self.cluster_(n_clusters=k).fit_predict(data[resampled_indices, :])

                connectivity_matrix += self.compute_iteration_connectivity_matrix(
                    resampled_indices, Mh, k, data.shape[0]
                )

                indicator_matrix += self.compute_iteration_indicator_mat(
                    resampled_indices, data.shape[0]
                )

            consensus_mats[k - self.min_k] = self.compute_consensus_mat(
                connectivity_matrix, indicator_matrix
            )

        self.consensus_matrices = consensus_mats

        self.bestK = self.compute_bestk()

    def compute_iteration_connectivity_matrix(
        self, resampled_indices, clust_res, k, nb_samples
    ):
        """
        Compute connectivity matrix

        The connectivity matrix allows to track how many times 2 elements of the matrix where sampled and clustered together

        Arguments
        ---------
        resampled_indices: array
            indices of the elements selected for the current iteration
        clust_res: array
            Clustering results
        k: int
            number of clusters
        nb_samples: int
            total number of elements of the input matrix


        Returns
        -------
        conn_mat
            Connectivity matrix for the iteration
        """
        # find indices of elements from same clusters with bisection
        # on sorted array => this is more efficient than brute force search
        conn_mat = np.zeros((nb_samples,) * 2)
        index_mapping = np.array((clust_res, resampled_indices)).T
        index_mapping = index_mapping[index_mapping[:, 0].argsort()]
        sorted_ = index_mapping[:, 0]
        id_clusts = index_mapping[:, 1]
        for i in range(k):  # for each cluster
            ia = bisect.bisect_left(sorted_, i)
            ib = bisect.bisect_right(sorted_, i)
            is_ = np.sort(id_clusts[ia:ib])
            ids_ = np.array(list(combinations(is_, 2))).T
            # sometimes only one element is in a cluster (no combinations)
            if ids_.size != 0:
                conn_mat[ids_[0], ids_[1]] += 1
        return conn_mat

    def compute_iteration_indicator_mat(self, resampled_indices, nb_samples):
        """
        Compute indicator matrix for one iteration

        The indicator matrix allows to track how many time 2 elements of the matrix where sampled together

        Arguments
        ---------
        resampled_indices: array
            sample's indices to use for the iteration
        nb_samples: int
            Number of samples in the initial matrix

        Returns
        -------
        indic_mat
            indicator matrix for the current iteration
        """
        indic_mat = np.zeros((nb_samples,) * 2)
        ids_2 = np.array(list(combinations(np.sort(resampled_indices), 2))).T
        indic_mat[ids_2[0], ids_2[1]] += 1
        return indic_mat

    def compute_consensus_mat(self, connectivity_mat, indicator_mat):
        """
        Compute consensus matrix defined as the normalized sum of the connectivity matrices of all the resampled datasets

        Arguments
        ---------
        connectivity_mat: ndarray
            connectivity matrix for k clusters. Sum of the iteration connectivity matrix
        indicator_mat: ndarray
            indicator matrix for k clusters. Sum of the iteration indicator matrix

        Returns
        -------
        consensus_mat
            Consensus matrix for k clusters
        """
        consensus_mat = connectivity_mat / (indicator_mat + 1e-8)
        # consensus_mat is upper triangular (with zeros on diagonal), we now make it symmetric
        consensus_mat += consensus_mat.T
        consensus_mat[
            range(consensus_mat.shape[0]), range(consensus_mat.shape[0])
        ] = 1  # always with self
        return consensus_mat

    def compute_bestk(self):
        """
        Get best number of clusters

        Returns
        -------
        best number of clusters
        """
        assert self.consensus_matrices is not None, "First compute consensus clustering"
        self.Ak = self.compute_area_under_curve()
        self.deltaK = self.compute_area_delta()
        return (
            np.argmax(self.deltaK) + self.min_k if self.deltaK.size > 0 else self.min_k
        )

    def compute_area_delta(self):
        """
        Compute the differences between areas under CDFs

        Returns
        -------
        Array containing the difference between the area under the CDFs for each number of clusters
        """
        assert self.consensus_matrices is not None, "First compute consensus clustering"
        return np.array(
            [
                (Ab - Aa) / Aa if i > 2 else Aa / self.nbins
                for Ab, Aa, i in zip(
                    self.Ak[1:], self.Ak[:-1], range(self.min_k, self.max_k + 1)
                )
            ]
        )

    def compute_area_under_curve(self):
        """
        Compute area under the CDFs curve

        Returns
        -------
        area_under_curve
            array of the area under the CDF for each cluster number
        """
        assert self.consensus_matrices is not None, "First compute consensus clustering"
        area_under_curve = np.zeros(self.max_k - self.min_k + 1)
        for i, m in enumerate(self.consensus_matrices):
            hist, bins = np.histogram(m.ravel(), density=True, bins=self.nbins)
            area_under_curve[i] = sum(
                h * (b - a) for b, a, h in zip(bins[1:], bins[:-1], np.cumsum(hist))
            )
        return area_under_curve

    def compute_summary_statistics(self, k):
        """
        For one prediction, compute a summary statistics, cluster consensus and item consensus, showing cluster stability and most representative cluster items.

        Cluster consensus is defined as the average consensus index between all pairs of items belonging to the cluster.
        Item consensus is defined as the average consensus index between item ei and all the (other) items in a cluster.


        Arguments
        ---------
        k: int
            Number of clusters

        Returns
        -------
        predictions
            Array of the predicted cluster
        clusters_consensus
            Array of cluster consensus
        items_consensus
            Array of the item consensus (nb_items * k)
        """
        assert self.consensus_matrices is not None, "First compute consensus clustering"
        assert (
            self.min_k <= k <= self.max_k
        ), "Number of clusters must be between min_k and max_k"
        predictions = self.predict(k)
        clusters_consensus = self.compute_clusters_consensus(predictions, k)
        items_consensus = self.compute_items_consensus(predictions, k)
        return predictions, clusters_consensus, items_consensus

    def compute_clusters_consensus(self, prediction, k):
        """
        For one prediction, compute clusters consensus, showing cluster stability.

        Cluster consensus is defined as the average consensus index between all pairs of items belonging to the cluster.

        Arguments
        ---------
        prediction: ndarray
            Array of the predicted cluster
        k: int
            Number of clusters

        Returns
        -------
        clusters_consensus
            Array of cluster consensus
        """
        assert (
            self.min_k <= k <= self.max_k
        ), "Number of clusters must be between min_k and max_k"
        clusters_consensus = np.zeros(k)
        for clust in range(k):
            ids = np.where(prediction == clust)[0]
            clust_size = len(ids)
            ids_ = np.array(list(combinations(np.sort(ids), 2))).T
            clusters_consensus[clust] = self.consensus_matrices[
                k - self.min_k, ids_[0], ids_[1]
            ].sum() / (clust_size * (clust_size - 1) / 2)
        return clusters_consensus

    def compute_items_consensus(self, prediction, k):
        """
        For one prediction, compute item consensus, showing most representative cluster items.

        Item consensus is defined as the average consensus index between item ei and all the (other) items in a cluster.

        Arguments
        ---------
        prediction: ndarray
            Array of the predicted cluster
        k: int
            Number of clusters

        Returns
        -------
        items_consensus
            Array of the item consensus (nb_items * k)
        """
        assert (
            self.min_k <= k <= self.max_k
        ), "Number of clusters must be between min_k and max_k"
        items_consensus = np.zeros(
            (self.consensus_matrices[k - self.min_k].shape[0], k)
        )

        clusters, sizes = np.unique(prediction, return_counts=True)

        for id in range(items_consensus.shape[0]):
            for clust, siz in zip(clusters, sizes):
                clust_elem = np.where(prediction == clust)[0]
                if id in clust_elem:
                    cols = clust_elem[clust_elem != id]
                    items_consensus[id, clust] = np.sum(
                        self.consensus_matrices[k - self.min_k, id, cols]
                    ) / (siz - 1)
                else:
                    items_consensus[id, clust] = (
                        np.sum(self.consensus_matrices[k - self.min_k, id, clust_elem])
                        / siz
                    )
        return items_consensus

    def predict(self, k):
        """
        Predicts clusters on the consensus matrix, for k clusters using the consensus matrix

        Arguments
        ---------
        k: int
            Number of clusters

        Returns
        -------
        predicted cluster for k clusters
        """
        assert self.consensus_matrices is not None, "First compute consensus clustering"
        assert (
            self.min_k <= k <= self.max_k
        ), "Number of clusters must be between min_k and max_k"
        return self.cluster_(n_clusters=k).fit_predict(
            1 - self.consensus_matrices[k - self.min_k]
        )

    def predict_data(self, data):
        """
        Predicts clusters on the data, for best found cluster number

        Arguments
        ---------
        data: ndarray
            input matrix (samples * attributes)

        Returns
        -------
        predicted cluster for best number of clusters
        """
        assert self.consensus_matrices is not None, "First compute consensus clustering"
        return self.cluster_(n_clusters=self.bestK).fit_predict(data)

    def plot_clustermap(self, k, saving_path, col_color=None):
        """
        Plot and save clustermap

        Arguments
        ---------
        k: int
            The number of cluster to plot the clustermap for
        saving_path: str
            Path to file where to store the result plot
        col_color: dataframe
            dataframe of color to annotate the clustermap.
            NOTE: the colors must be ordered in the same way as the initial data matrix and be in hexadecimal or rgb format (cf clustermap seaborn documentation)

        """
        assert self.consensus_matrices is not None, "First compute consensus clustering"
        assert (
            self.min_k <= k <= self.max_k
        ), "Number of clusters must be between min_k and max_k"
        cluster_map = sn.clustermap(
            pd.DataFrame(self.consensus_matrices[k - self.min_k]),
            method="ward",
            metric="euclidean",
            cmap=sn.color_palette("Blues", as_cmap=True),
            yticklabels=False,
            xticklabels=False,
            figsize=(10, 10),
            col_colors=col_color,
        )
        cluster_map.figure.savefig(saving_path)

    def build_clusters_consensus_df(self):
        """
        Compute cluster consensus for each k from min_k to max_k and return a dataframe to use in the plot_clusters_consensus
        """
        consensus_clusters = []
        for k in range(self.min_k, self.max_k + 1):
            prediction = self.predict(k)
            consensus_clust = self.compute_clusters_consensus(prediction, k)
            consensus_clusters.append(consensus_clust)

        index_name = [f"k={i}" for i in range(self.min_k, self.max_k + 1)]

        return pd.DataFrame(consensus_clusters, index=index_name)

    def plot_clusters_consensus(self, cons_clust_df, fig_path):
        """
        Plot cluster consensus results showing the statibility of clusters

        To plot cluster consensus you first need to compute clusters consensus
        for each K between mi_k and max_k.

        Arguments
        ---------
        cons_clust_df: dataframe
            clusters consensus results formatted in a dataframe max_k-min_k+1*max_k
        fig_path: str
            Path to store the plot
        """
        fig, ax = plt.subplots(figsize=(13, 8))

        cons_clust_df.plot(
            kind="bar",
            stacked=False,
            width=1,
            edgecolor="white",
            linewidth=1,
            align="center",
            ax=ax,
        )
        plt.tick_params(rotation=45)
        plt.xlabel("Number of clusters")
        plt.ylabel("Cluster consensus measuring cluster stability")
        ax.legend(bbox_to_anchor=(1, 1), title="Cluster")
        plt.savefig(fig_path)

    def line_plots_cluster_consensus(self, cons_clust_df, fig_path, threshold=0.75):
        """
        Line plots of the % and number of cluster with cluster consensus > threshold

        To plot cluster consensus you first need to compute clusters consensus
        for each K between mi_k and max_k.

        Arguments
        -----------
        cons_clust_df: dataframe
            clusters consensus results formatted in a dataframe max_k-min_k+1*max_k
        fig_path: str
            Path to store the plot
        threshold: float
            cluster consensus minimum value to be considered as stable
            default=0.75
        """
        ratio_stable_cluster = cons_clust_df.apply(
            lambda x: (x > threshold).sum() / (~x.isna()).sum(), axis=1
        )
        nb_stable_cluster = cons_clust_df.apply(lambda x: (x > threshold).sum(), axis=1)

        fig, (ax1, ax2) = plt.subplots(2)
        sn.lineplot(nb_stable_cluster, ax=ax1)
        ax1.set_title("Cluster stability")
        ax1.set(ylabel=f"number of cluster \n with stability>{threshold}", xlabel=" ")
        sn.lineplot(ratio_stable_cluster, ax=ax2, color="orange")
        ax2.set(
            ylabel=f"% of cluster \n with stability>{threshold}",
            xlabel="total nb of cluster",
        )
        plt.savefig(fig_path)

    def plot_deltak(self, fig_path):
        """
        Plot the curve showing the relative change in area under the curve

        Arguments
        ---------
        fig_path: str
            Path to store the plot
        """
        fig, ax = plt.subplots()
        plt.plot(range(self.min_k, self.max_k), self.deltaK, marker="o", color="black")
        plt.xlabel("Number of clusters")
        plt.ylabel("Relative change in area under the curve")
        plt.savefig(fig_path)



import unittest
import numpy as np

from inmoose.deseq2 import (
    DESeqDataSet,
    makeExampleDESeqDataSet,
    estimateDispersionsGeneEst,
    estimateDispersionsFit,
    estimateDispersionsMAP,
    nbinomWaldTest,
)


class Test(unittest.TestCase):
    def test_linear_mu(self):
        """test that the use of linear model for fitting mu works as expected"""

        dds = makeExampleDESeqDataSet(
            n=100,
            m=4,
            interceptMean=10,
            interceptSD=3,
            dispMeanRel=lambda x: 0.5,
            sizeFactors=[0.5, 1, 1, 2],
        )
        dds = dds.estimateSizeFactors()
        dds1 = DESeqDataSet(dds.copy())
        dds1 = estimateDispersionsGeneEst(dds1, linearMu=False)
        dds2 = estimateDispersionsGeneEst(dds.copy(), linearMu=True)
        mu1 = dds1.layers["mu"]
        mu2 = dds2.layers["mu"]
        cors = np.diag(np.corrcoef(mu1, mu2)[: mu1.shape[0], mu1.shape[1] :])
        self.assertTrue(np.all(cors > 1 - 1e-6))

        dds2 = estimateDispersionsFit(dds2, fitType="mean")
        dds2 = estimateDispersionsMAP(dds2)
        dds2 = nbinomWaldTest(dds2)
        res = dds2.results()



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/estimateGLMTagwiseDisp.R' of the Bioconductor edgeR package (version 3.38.4).


import logging
import numpy as np

from .aveLogCPM import aveLogCPM
from .dispCoxReidInterpolateTagwise import dispCoxReidInterpolateTagwise


def estimateGLMTagwiseDisp_DGEList(
    self, design=None, prior_df=10, trend=None, span=None
):
    """
    Compute an empirical Bayes estimate of the negative binomial dispersion
    parameter for each tag, with expression levels specified by a log-linear
    model.

    NB: Modifies :code:`self` in place

    Arguments
    ---------
    self : DGEList
    design : matrix
        the design matrix, as in :code:`glmFit`
    prior_df : int
        prior degrees of freedom
    trend : bool
        whether the prior should be the trended (:code:`trend=True`) or the
        common dispersion (:code:`trend=False`)
    span : float, optional
        width of the smoothing window, in terms of proportion of the data set.
        Default value decreases with the number of tags.

    Returns
    -------
    DGEList
        the input :code:`self`, updated with the tagwise dispersion parameter
        estimate for each tag for the negative binomial model that maximizes the
        Cox-Reid adjusted profile likelihood.
    """
    # Find appropriate dispersion
    if trend is None:
        trend = self.trended_dispersion is not None
    if trend:
        dispersion = self.trended_dispersion
        if dispersion is None:
            raise ValueError(
                "No trended_dispersion in data object. Run estimateGLMTrendedDisp first."
            )
    else:
        dispersion = self.common_dispersion
        if dispersion is None:
            raise ValueError(
                "No common_dispersion found in data object. Run estimateGLMCommonDisp first."
            )

    if self.AveLogCPM is None:
        self.AveLogCPM = self.aveLogCPM()
    ntags = self.counts.shape[0]

    if span is None:
        if ntags > 10:
            span = (10 / ntags) ** 0.23
        else:
            span = 1
    self.span = span

    d = estimateGLMTagwiseDisp(
        self.counts,
        design=design,
        offset=self.getOffset(),
        dispersion=dispersion,
        trend=trend,
        span=span,
        prior_df=prior_df,
        AveLogCPM=self.AveLogCPM,
        weights=self.weights,
    )
    self.prior_df = prior_df
    self.tagwise_dispersion = d
    return self


def estimateGLMTagwiseDisp(
    y,
    dispersion,
    design=None,
    offset=None,
    prior_df=10,
    trend=True,
    span=None,
    AveLogCPM=None,
    weights=None,
):
    """
    Compute an empirical Bayes estimate of the negative binomial dispersion
    parameter for each tag, with expression levels specified by a log-linear
    model.

    This function implements the empirical Bayes strategy propose by McCarthy et
    al. (2012) [2]_ for estimating the tagwise negative binomial dispersions.
    The experimental conditions are specified by design matrix allowing for
    multiple explanatory factors. The empirical Bayes posterior is implemented
    as a conditional likelihood with tag-specific weights, and the conditional
    likelihood is computed using Cox-Reid approximate conditional likelihood
    (Cox and Reid, 1987 [1]_).

    The prior degrees of freedom determine the weight given to the global
    dispersion trend. The larger the prior degrees of freedom, the more the
    tagwise dispersions are squeezed towards the global trend.

    Note that the terms "tag" and "gene" are synonymous here. The function is
    only named "tagwise" for historical reasons.

    This function calls the lower-level :func:`dispCoxReidInterpolateTagwise`.

    Arguments
    ---------
    y : matrix
        matrix of counts
    dispersion : float or array_like
        common or trended dispersion estimates, used as an initial estimate for
        the tagwise estimates
    design : matrix
        the design matrix, as in :code:`glmFit`
    offset : matrix, optional
        offset matrix for the log-linear model, as in :func:`glmFit`. Defaults
        to the log-effective library sizes.
    prior_df : int
        prior degrees of freedom
    trend : bool
        whether the prior should be the trended (:code:`trend=True`) or the
        common dispersion (:code:`trend=False`)
    span : float, optional
        width of the smoothing window, in terms of proportion of the data set.
        Default value decreases with the number of tags.
    AveLogCPM : array_like, optional
        vector of the average log2 counts per million for each tag
    weights : matrix, optional
        observation weights

    Returns
    -------
    ndarray
        vector of the tagwise dispersion estimates

    References
    ----------
    .. [1] D. R. Cox, N. Reid. 1987. Parameter orthogonality and approximate
       conditional inference. Journal of the Royal Statistical Society Series B
       49, 1-39.
    .. [2] D. J. McCarthy, Y. Chen, G. K. Smyth. 2012. Differential expression
       analysis of multifactor RNA-Seq experiments with respect to biological
       variation. Nucleic Acids Research 40, 4288-4297. :doi:`10.1093/nar/gks042`
    """

    # Check y
    y = np.asarray(y, order="F")
    (ntags, nlibs) = y.shape
    if ntags == 0:
        return 0

    # Check design
    if design is None:
        design = np.ones((y.shape[1], 1), order="F")
    else:
        design = np.asarray(design, order="F")

    if design.shape[1] >= y.shape[1]:
        logging.warnings.warn("No residual df: setting dispersion to NA")
        return np.full((ntags,), np.nan, order="F")

    # Check offset
    if offset is None:
        offset = np.log(y.sum(axis=0))

    # Check span
    # span can be chosen smaller when ntags is large
    if span is None:
        if ntags > 10:
            span = (10 / ntags) ** 0.23
        else:
            span = 1

    # Check AveLogCPM
    if AveLogCPM is None:
        AveLogCPM = aveLogCPM(y, offset=offset, weights=weights)

    # Call Cox-Reid grid method
    tagwise_dispersion = dispCoxReidInterpolateTagwise(
        y,
        design,
        offset=offset,
        dispersion=dispersion,
        trend=trend,
        prior_df=prior_df,
        span=span,
        AveLogCPM=AveLogCPM,
        weights=weights,
    )
    return tagwise_dispersion



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------


class DGEGLM(object):
    def __init__(self, fit):
        # fit is a 5-tuple
        (coefficients, fitted_values, deviance, iter, failed) = fit
        self.coefficients = coefficients
        self.fitted_values = fitted_values
        self.deviance = deviance
        self.iter = iter
        self.failed = failed

        self.counts = None
        self.design = None
        self.offset = None
        self.dispersion = None
        self.weights = None
        self.prior_count = None
        self.unshrunk_coefficients = None
        self.method = None



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/splitIntoGroups.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
from ..utils import asfactor


def splitIntoGroups_DGEList(self):
    """
    Split the counts according to group

    See also
    --------
    splitIntoGroups

    Arguments
    ---------
    self : DGEList
        DGEList object, from which the matrix of counts is taken
    group : array_like or :obj:`Factor`
        vector or factor giving the experimental group/condition for each
        library

    Returns
    -------
    list
        list in which each element is a matrix of counts for an individual group.
    """
    group = self.samples.group
    return splitIntoGroups(self.counts, group=group)


def splitIntoGroups(y, group=None):
    """
    Split the counts according to group

    Split the counts from a matrix of counts according to a group, creating a
    list where each element consists of a numeric matrix of counts for a
    particular experimental group.

    Arguments
    ---------
    y : array_like
        matrix of counts
    group : array_like or :obj:`Factor`
        vector or factor giving the experimental group/condition for each
        library

    Returns
    -------
    list
        list in which each element is a matrix of counts for an individual group.
    """
    # Check y
    (ntags, nlibs) = y.shape

    # Check group
    if group is None:
        group = np.ones(nlibs)
    if len(group) != nlibs:
        raise ValueError("Incorrect length of group.")
    group = asfactor(group).droplevels()

    out = [y.T[group == i].T for i in np.unique(group)]
    return out



import unittest
import numpy as np

from inmoose.edgepy import is_integer_array


class test_objects(unittest.TestCase):
    def test_is_integer(self):
        arr = np.array([0, 1, 2], dtype="int")
        self.assertEqual(is_integer_array(arr), True)

        arr = np.array([0, 1, 2], dtype="double")
        self.assertEqual(is_integer_array(arr), False)

        arr = np.array([0, 1, 2], dtype="str")
        with self.assertRaisesRegex(
            RuntimeError, expected_regex="array dtype is neither 'int' nor 'double'"
        ):
            is_integer_array(arr)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np

from inmoose.utils import Factor
from inmoose.deseq2 import makeExampleDESeqDataSet, DESeq, nbinomWaldTest


class Test(unittest.TestCase):
    def test_addMLE(self):
        """test that adding MLE works as expected"""
        dds = makeExampleDESeqDataSet(n=200, m=12, betaSD=1, seed=42)
        dds.obs["condition"] = Factor(np.repeat(["A", "B", "C"], 4))
        dds.design = "~condition"
        dds = DESeq(dds, betaPrior=True)
        ddsNP = nbinomWaldTest(dds.copy(), betaPrior=False)

        res1 = dds.results(contrast=["condition", "C", "A"], addMLE=True)
        res2 = ddsNP.results(contrast=["condition", "C", "A"])
        self.assertTrue(np.allclose(res1.lfcMLE, res2.log2FoldChange, equal_nan=True))

        res1 = dds.results(contrast=["condition", "A", "B"], addMLE=True)
        res2 = ddsNP.results(contrast=["condition", "A", "B"])
        self.assertTrue(np.allclose(res1.lfcMLE, res2.log2FoldChange, equal_nan=True))

        res1 = dds.results(contrast=["condition", "C", "B"], addMLE=True)
        res2 = ddsNP.results(contrast=["condition", "C", "B"])
        self.assertTrue(np.allclose(res1.lfcMLE, res2.log2FoldChange, equal_nan=True))



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 Yuqing Zhang
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/ComBat_seq.R' of the Bioconductor sva package (version 3.44.0).

import logging
import numpy as np
import pandas as pd

from ..edgepy import DGEList, estimateGLMCommonDisp, estimateGLMTagwiseDisp, glmFit
from ..utils import asfactor
from .covariates import make_design_matrix
from .helper_seq import vec2mat, match_quantiles


def pycombat_seq(
    counts,
    batch,
    covar_mod=None,
    shrink=False,
    shrink_disp=False,
    gene_subset_n=None,
    ref_batch=None,
    na_cov_action="raise",
):
    """pycombat_seq is an improved model from ComBat using negative binomial regression, which specifically targets RNA-Seq count data.

    Arguments
    ---------
    counts : matrix
        raw count matrix (dataframe or numpy array) from genomic studies (dimensions gene x sample)
    batch : array or list or :obj:`inmoose.utils.factor.Factor`
        Batch indices. Must have as many elements as the number of columns in the expression matrix.
    covar_mod : list or matrix, optional
        model matrix (dataframe, list or numpy array) for one or multiple covariates to include in linear model (signal
        from these variables are kept in data after adjustment). Covariates have to be categorial,
        they can not be continious values (default: `None`).
    shrink : bool, optional
        whether to apply shrinkage on parameter estimation
    shrink_disp : bool, optional
        whether to apply shrinkage on dispersion
    gene_subset_n : int, optional
        number of genes to use in emprirical Bayes estimation, only useful when shrink = True
    ref_batch : any, optional
        batch id of the batch to use as reference (default: `None`)
    na_cov_action : str
        Option to choose the way to handle missing covariates

        - :code:`"raise"` raise an error if missing covariates and stop the code
        - :code:`"remove"` remove samples with missing covariates and raise a warning
        - :code:`"fill"` handle missing covariates, by creating a distinct
          covariate per batch
        (default: :code:`"raise"`)

    Returns
    -------
    matrix
        the input expression matrix adjusted for batch effects.
        same type as the input `data`
    """

    ####### Preparation #######
    # Handle batches, covariates and prepare design matrix
    vci = make_design_matrix(
        counts,
        batch,
        covar_mod=covar_mod,
        ref_batch=ref_batch,
        na_cov_action=na_cov_action,
    )

    counts = vci.counts
    list_samples = vci.list_samples
    list_genes = vci.list_genes
    batch = vci.batch
    design = vci.design
    batchmod = vci.batch_mod
    mod = vci.covar_mod
    ref_batch_index = vci.ref_batch_idx

    n_batch = vci.n_batch
    batches_ind = vci.batch_composition
    batch_sizes = {b: len(v) for b, v in batches_ind.items()}

    n_sample = counts.shape[1]

    # Remove genes with only 0 counts in any batch
    keep = np.full((counts.shape[0],), True)
    for b in batch.categories:
        # force ndarray instead of matrices (if any), and squeeze to remove dims of length 1
        keep &= np.asarray(counts[:, batch == b].sum(axis=1) > 0).squeeze()
    rm = np.logical_not(keep).nonzero()[0]
    keep = keep.nonzero()[0]
    countsOri = counts.copy()
    counts = counts[keep, :]

    dge_obj = DGEList(counts=counts)

    ####### Estimate gene-wise dispersions within each batch #######
    logging.info("Estimating dispersions")

    # Estimate common dispersion within each batch as an initial value
    def disp_common_helper(i):
        if (
            batch_sizes[i] <= design.shape[1] - batchmod.shape[1] + 1
            or np.linalg.matrix_rank(mod[batches_ind[i]]) < mod.shape[1]
        ):
            # not enough residual degree of freedom

            return estimateGLMCommonDisp(
                counts[:, batches_ind[i]], design=None, subset=counts.shape[0]
            )
        else:
            return estimateGLMCommonDisp(
                counts[:, batches_ind[i]],
                design=mod[batches_ind[i]],
                subset=counts.shape[0],
            )

    disp_common = {b: disp_common_helper(b) for b in batch.categories}

    # Estimate gene-wise dispersion within each batch
    def genewise_disp_helper(i):
        if (
            batch_sizes[i] <= design.shape[1] - batchmod.shape[1] + 1
            or np.linalg.matrix_rank(mod[batches_ind[i]]) < mod.shape[1]
        ):
            return [disp_common[i] for j in range(counts.shape[0])]
        else:
            return estimateGLMTagwiseDisp(
                counts[:, batches_ind[i]],
                design=mod[batches_ind[i]],
                dispersion=disp_common[i],
                prior_df=0,
            )

    genewise_disp_lst = {b: genewise_disp_helper(b) for b in batch.categories}
    # construction dispersion matrix
    phi_matrix = np.full(counts.shape, np.nan)
    for b in batch.categories:
        phi_matrix[:, batches_ind[b]] = vec2mat(genewise_disp_lst[b], batch_sizes[b])

    ####### Estimate parameters from NB GLM #######
    logging.info("Fitting the GLM model")
    # no intercept - nonEstimable; compute offset (library sizes) within function
    glm_f = dge_obj.glmFit(design=design, dispersion=phi_matrix, prior_count=1e-4)
    # compute intercept as batch-size-weighted average from batches
    alpha_g = glm_f.coefficients[:, range(n_batch)] @ (
        np.array([batch_sizes[batch.categories[i]] for i in range(n_batch)]) / n_sample
    )
    # original offset - sample (library size)
    new_offset = vec2mat(dge_obj.getOffset(), counts.shape[0]).T
    # new offset - gene background expression (dge_obj.getOffset() is the same as log(dge_obj.samples.lib_size)
    new_offset += vec2mat(alpha_g, counts.shape[1])

    glm_f2 = glmFit(
        dge_obj.counts,
        design=design,
        dispersion=phi_matrix,
        offset=new_offset,
        prior_count=1e-4,
    )

    gamma_hat = glm_f2.coefficients[:, range(n_batch)]
    mu_hat = glm_f2.fitted_values
    phi_hat = np.column_stack(
        [genewise_disp_lst[batch.categories[i]] for i in range(n_batch)]
    )

    ####### In each batch, compute posterior estimation through Monte-Carlo integration #######
    if shrink:
        logging.info("Apply shrinkage - computing posterior estimates for parameters")
        raise NotImplementedError
    else:
        logging.info("shrinkage off - using GLM estimates for parameters")
        gamma_star_mat = gamma_hat
        phi_star_mat = phi_hat

    ####### Obtain adjusted batch-free distribution #######
    mu_star = np.full(counts.shape, np.nan)
    for jj in range(n_batch):
        b = batch.categories[jj]
        mu_star[:, batches_ind[b]] = np.exp(
            np.log(mu_hat[:, batches_ind[b]])
            - vec2mat(gamma_star_mat[:, jj], batch_sizes[b])
        )
    phi_star = phi_star_mat.mean(axis=1)

    ####### Ajust the data #######
    logging.info("Adjusting the data")
    adjust_counts = np.full(counts.shape, np.nan)
    for kk in range(n_batch):
        b = batch.categories[kk]
        counts_sub = counts[:, batches_ind[b]]
        if kk == ref_batch_index:
            adjust_counts[:, batches_ind[b]] = counts_sub
            continue
        old_mu = mu_hat[:, batches_ind[b]]
        old_phi = phi_hat[:, kk]
        new_mu = mu_star[:, batches_ind[b]]
        new_phi = phi_star
        adjust_counts[:, batches_ind[b]] = match_quantiles(
            counts_sub=counts_sub,
            old_mu=old_mu,
            old_phi=old_phi,
            new_mu=new_mu,
            new_phi=new_phi,
        )

    # Add back genes with only 0 counts in any batch (so that dimensions do not change)
    adjust_counts_whole = np.full(countsOri.shape, np.nan)
    adjust_counts_whole[keep, :] = adjust_counts
    adjust_counts_whole[rm, :] = countsOri[rm, :]

    if isinstance(counts, pd.DataFrame):
        return pd.DataFrame(adjust_counts_whole, columns=list_samples, index=list_genes)
    else:
        return adjust_counts_whole



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# DESeq2 C++ functions ported back to Python.
# DESeq2 C++ functions use Armadillo for linear algebra operations (matrices
# and vectors indexing and slicing, matrices and vectors multiplication...)
# These capabilities are in fact performed in C directly by numpy, so we
# figure that there is no need to use Armadillo here, nor to remain full C++.
# This may change in the future, e.g. for performance reasons.
#
# Note: the canonical, up-to-date DESeq2.cpp lives in the DESeq2 library, the
# development branch of which can be viewed here:
# https://github.com/mikelove/DESeq2/blob/master/src/DESeq2.cpp

import numpy as np
import pandas as pd
from scipy.special import loggamma as lgamma
from scipy.special import digamma, polygamma
from scipy.special import xlog1py, xlogy

from ..utils import dnbinom_mu


def log_posterior(
    log_alpha,
    y,
    mu,
    x,
    log_alpha_prior_mean,
    log_alpha_prior_sigmasq,
    usePrior,
    weights,
    useWeights,
    weightThreshold,
    useCR,
):
    """
    This function returns the log posterior of dispersion parameter alpha, for
    negative binomial variables.  Given the counts :code:`y`, the expected
    means :code:`mu`, the design matrix :code:`x` (used for calculating the
    Cox-Reid adjustment), and the parameters for the normal prior on
    :code:`log_alpha`.

    Arguments
    ---------
    log_alpha : ndarray
        log of the dispersion parameters alpha, last dim is M (or broadcastable)
    y : ndarray
        counts, matrix of shape (N,M) or vector of length N
    mu : ndarray
        expected means, same shape as :code:`y`
    x : ndarray
        design matrix, shape (N,P)
    log_alpha_prior_mean : ndarray
        normal prior on log alpha, vector of length M
    log_alpha_prior_sigmasq : float
        standard deviation of the prior on log alpha
    usePrior : bool
        whether to use prior regularization
    weights : ndarray
        weights to apply, same shape as :code:`y`
    useWeights : bool
        whether to use weights
    weightThreshold : float
        minimal weight to consider in Cox-Reid regularization
    useCR : bool
        whether to use Cox-Reid regularization

    Returns
    -------
    ndarray
        the log value of the posterior, weighted and regularized as specified
        by the input arguments
        same shape as first argument :code:`log_alpha`
    """

    # make sure that all arrays are broadcastable to the appropriate shapes
    if not isinstance(log_alpha, np.ndarray):
        log_alpha = np.repeat(log_alpha, 1)
    if len(y.shape) == 1:
        y = y[:, None]
    if len(mu.shape) == 1:
        mu = mu[:, None]
    if len(weights.shape) == 1:
        weights = weights[:, None]

    # helper variables to control the shapes
    M = np.maximum(y.shape[-1], log_alpha.shape[-1])
    (N, P) = x.shape

    alpha = np.exp(log_alpha)
    if useCR:
        x = x[None]
        # NB: now, x.shape == (1, N, P)
        mu_neg1 = 1.0 / mu
        w_diag = 1.0 / (mu_neg1 + np.expand_dims(alpha, axis=-2))
        if useWeights:
            # cancel out all weights below the threshold
            idx = weights <= weightThreshold
            w_diag[np.broadcast_to(idx, w_diag.shape)] = 0.0
        assert w_diag.shape[-2:] == (N, M)
        assert w_diag.shape[:-2] == log_alpha.shape[:-1]

        # use `np.swapaxes` to transpose the matrices stored in the last 2 dims
        w_diag = np.swapaxes(w_diag, -1, -2)
        # insert a new axis in last position
        w_diag = np.expand_dims(w_diag, axis=-1)
        assert w_diag.shape[-3:] == (M, N, 1)
        assert w_diag.shape[:-3] == log_alpha.shape[:-1]
        b = np.swapaxes(x * w_diag, -1, -2) @ x
        assert b.shape[-3:] == (M, P, P)
        assert b.shape[:-3] == log_alpha.shape[:-1]
        cr_term = -0.5 * np.linalg.slogdet(b)[1]
        assert cr_term.shape[:-1] == log_alpha.shape[:-1]
        assert cr_term.shape[-1] == M
    else:
        cr_term = 0.0

    # insert a new axis before the last one, to broadcast on the N dimension
    # of y, mu, weights
    alpha = np.expand_dims(alpha, axis=-2)
    alpha_neg1 = 1.0 / alpha
    if useWeights:
        ll_part = np.sum(
            weights
            * (
                lgamma(y + alpha_neg1)
                - lgamma(alpha_neg1)
                - xlogy(y, mu + alpha_neg1)
                - xlog1py(alpha_neg1, alpha * mu)
            ),
            axis=-2,
        )
    else:
        ll_part = np.sum(
            lgamma(y + alpha_neg1)
            - lgamma(alpha_neg1)
            - xlogy(y, mu + alpha_neg1)
            - xlog1py(alpha_neg1, alpha * mu),
            axis=-2,
        )

    assert (
        ll_part.shape[:-1] == log_alpha.shape[:-1]
    ), f"{ll_part.shape} vs {log_alpha.shape}"
    assert ll_part.shape[-1] == M

    if usePrior:
        prior_part = (
            -0.5 * (log_alpha - log_alpha_prior_mean) ** 2 / log_alpha_prior_sigmasq
        )
        assert (
            prior_part.shape[:-1] == log_alpha.shape[:-1]
        ), f"{prior_part.shape} vs {log_alpha.shape}"
    else:
        prior_part = 0.0

    return ll_part + prior_part + cr_term


def dlog_posterior(
    log_alpha,
    y,
    mu,
    x,
    log_alpha_prior_mean,
    log_alpha_prior_sigmasq,
    usePrior,
    weights,
    useWeights,
    weightThreshold,
    useCR,
):
    """
    This function returns the derivative of the log posterior with respect to
    the log of the dispersion parameter alpha, given the same inputs as
    :func:`log_posterior`.

    Arguments
    ---------
    log_alpha : ndarray
        log of the dispersion parameters alpha, last dim is M (or broadcastable)
    y : ndarray
        counts, matrix of shape (N,M) or vector of length N
    mu : ndarray
        expected means, same shape as :code:`y`
    x : ndarray
        design matrix, shape (N,P)
    log_alpha_prior_mean : ndarray
        normal prior on log alpha, vector of length M
    log_alpha_prior_sigmasq : float
        standard deviation of the prior on log alpha
    usePrior : bool
        whether to use prior regularization
    weights : ndarray
        weights to apply, same shape as :code:`y`
    useWeights : bool
        whether to use weights
    weightThreshold : float
        minimal weight to consider in Cox-Reid regularization
    useCR : bool
        whether to use Cox-Reid regularization

    Returns
    -------
    ndarray
        the derivative of the log value of the posterior, weighted and
        regularized as specified by the input arguments
        same shape as first argument :code:`log_alpha`
    """

    # make sure that all arrays are broadcastable to the appropriate shapes
    if not isinstance(log_alpha, np.ndarray):
        log_alpha = np.repeat(log_alpha, 1)
    if len(y.shape) == 1:
        y = y[:, None]
    if len(mu.shape) == 1:
        mu = mu[:, None]
    if len(weights.shape) == 1:
        weights = weights[:, None]

    # helper variables to control the shapes
    M = log_alpha.shape[0]
    (N, P) = x.shape

    alpha = np.exp(log_alpha)
    if useCR:
        x = x[None]
        # NB: now, x.shape == (1, N, P)
        mu_neg1 = 1.0 / mu
        w_diag = 1.0 / (mu_neg1 + alpha[None])
        dw_diag = -np.power(mu_neg1 + alpha[None], -2)
        assert w_diag.shape == (N, M)
        assert dw_diag.shape == (N, M)
        # NB: w_diag.shape == dw_diag.shape == mu.shape == (N, M)
        if useWeights:
            # cancel out all weights below the threshold
            idx = weights <= weightThreshold
            w_diag[np.broadcast_to(idx, w_diag.shape)] = 0.0
            dw_diag[np.broadcast_to(idx, dw_diag.shape)] = 0.0

        # use `np.swapaxes` to transpose the matrices stored in the last 2 dims
        w_diag = np.swapaxes(w_diag, -1, -2)
        dw_diag = np.swapaxes(dw_diag, -1, -2)
        # insert a new axis in last position
        w_diag = np.expand_dims(w_diag, axis=-1)
        dw_diag = np.expand_dims(dw_diag, axis=-1)
        b = np.swapaxes(x * w_diag, -1, -2) @ x
        db = np.swapaxes(x * dw_diag, -1, -2) @ x
        assert b.shape == (M, P, P)
        assert db.shape == (M, P, P)
        cr_term = -0.5 * np.trace(np.linalg.inv(b) @ db, axis1=-2, axis2=-1)
        # NB original code computes
        #   ddetb = det(b) * trace(b.i() * db)
        # then
        #   cr_term = -0.5 * ddetb / det(b)
        # not sure why they multiply/divide by det(b)...
        assert cr_term.shape == alpha.shape, f"{cr_term.shape} vs {alpha.shape}"
    else:
        cr_term = 0.0

    alpha = alpha[None]
    alpha_neg1 = 1.0 / alpha
    alpha_neg2 = np.power(alpha, -2)
    alphamu = alpha * mu
    if useWeights:
        ll_part = alpha_neg2.squeeze() * np.sum(
            weights
            * (
                digamma(alpha_neg1)
                + np.log(1 + alphamu)
                - alphamu / (1.0 + alphamu)
                - digamma(y + alpha_neg1)
                + y / (mu + alpha_neg1)
            ),
            axis=0,
        )
    else:
        ll_part = alpha_neg2.squeeze() * np.sum(
            digamma(alpha_neg1)
            + np.log(1 + alphamu)
            - alphamu / (1.0 + alphamu)
            - digamma(y + alpha_neg1)
            + y / (mu + alpha_neg1),
            axis=0,
        )

    # only the prior part is wrt log alpha
    if usePrior:
        prior_part = -1.0 * (log_alpha - log_alpha_prior_mean) / log_alpha_prior_sigmasq
    else:
        prior_part = 0.0

    # note: return dlog_post / dalpha * alpha because we take derivatives wrt log alpha
    return (ll_part + cr_term) * alpha.squeeze() + prior_part


def d2log_posterior(
    log_alpha,
    y,
    mu,
    x,
    log_alpha_prior_mean,
    log_alpha_prior_sigmasq,
    usePrior,
    weights,
    useWeights,
    weightThreshold,
    useCR,
):
    """
    This function returns the second derivative of the log posterior with
    respect to the log of the dispersion parameter alpha, given the same inputs
    as :func:`log_posterior`.

    Arguments
    ---------
    log_alpha : ndarray
        log of the dispersion parameters alpha, last dim is M (or broadcastable)
    y : ndarray
        counts, matrix of shape (N,M) or vector of length N
    mu : ndarray
        expected means, same shape as :code:`y`
    x : ndarray
        design matrix, shape (N,P)
    log_alpha_prior_mean : ndarray
        normal prior on log alpha, vector of length M
    log_alpha_prior_sigmasq : float
        standard deviation of the prior on log alpha
    usePrior : bool
        whether to use prior regularization
    weights : ndarray
        weights to apply, same shape as :code:`y`
    useWeights : bool
        whether to use weights
    weightThreshold : float
        minimal weight to consider in Cox-Reid regularization
    useCR : bool
        whether to use Cox-Reid regularization

    Returns
    -------
    ndarray
        the second derivative of the log value of the posterior, weighted and
        regularized as specified by the input arguments
        same shape as first argument :code:`log_alpha`
    """

    # make sure that all arrays are broadcastable to the appropriate shapes
    if not isinstance(log_alpha, np.ndarray):
        log_alpha = np.repeat(log_alpha, 1)
    if len(y.shape) == 1:
        y = y[:, None]
    if len(mu.shape) == 1:
        mu = mu[:, None]
    if len(weights.shape) == 1:
        weights = weights[:, None]

    # helper variables to control the shapes
    M = log_alpha.shape[0]
    (N, P) = x.shape

    alpha = np.exp(log_alpha)
    if useCR:
        x = x[None]
        # NB: now, x.shape == (1, N, P)
        mu_neg1 = 1.0 / mu
        w_diag = 1.0 / (mu_neg1 + alpha[None])
        dw_diag = -1.0 * np.power(mu_neg1 + alpha[None], -2)
        d2w_diag = 2.0 * np.power(mu_neg1 + alpha[None], -3)
        assert w_diag.shape == (N, M)
        assert dw_diag.shape == (N, M)
        assert d2w_diag.shape == (N, M)
        # NB: w_diag.shape == dw_diag.shape == d2w_diag.shape == mu.shape == (N, M)
        if useWeights:
            # cancel out all weights below the threshold
            idx = weights <= weightThreshold
            w_diag[np.broadcast_to(idx, w_diag.shape)] = 0.0
            dw_diag[np.broadcast_to(idx, dw_diag.shape)] = 0.0
            d2w_diag[np.broadcast_to(idx, d2w_diag.shape)] = 0.0

        # use `np.swapaxes` to transpose the matrices stored in the last 2 dims
        w_diag = np.swapaxes(w_diag, -1, -2)
        dw_diag = np.swapaxes(dw_diag, -1, -2)
        d2w_diag = np.swapaxes(d2w_diag, -1, -2)
        # insert a new axis in last position
        w_diag = np.expand_dims(w_diag, axis=-1)
        dw_diag = np.expand_dims(dw_diag, axis=-1)
        d2w_diag = np.expand_dims(d2w_diag, axis=-1)
        b = np.swapaxes(x * w_diag, -1, -2) @ x
        b_i = np.linalg.inv(b)
        db = np.swapaxes(x * dw_diag, -1, -2) @ x
        d2b = np.swapaxes(x * d2w_diag, -1, -2) @ x
        assert b.shape == (M, P, P)
        assert db.shape == (M, P, P)
        assert d2b.shape == (M, P, P)

        ddetb = np.trace(b_i @ db, axis1=-2, axis2=-1)
        d2detb = (
            np.power(ddetb, 2)
            - np.trace(b_i @ db @ b_i @ db, axis1=-2, axis2=-1)
            + np.trace(b_i @ d2b, axis1=-2, axis2=-1)
        )
        cr_term = 0.5 * np.power(ddetb, 2) - 0.5 * d2detb
        assert cr_term.shape == alpha.shape, f"{cr_term.shape} vs {alpha.shape}"
        x = x.squeeze(-3)
    else:
        cr_term = 0.0

    alpha = alpha[None]
    alpha_neg1 = 1.0 / alpha
    alpha_neg2 = np.power(alpha, -2)
    alphamu = alpha * mu
    if useWeights:
        ll_part = -2 * np.power(alpha, -3) * np.sum(
            weights
            * (
                digamma(alpha_neg1)
                + np.log(1 + alphamu)
                - alphamu / (1 + alphamu)
                - digamma(y + alpha_neg1)
                + y / (mu + alpha_neg1)
            )
        ) + alpha_neg2 * np.sum(
            weights
            * (
                -1 * alpha_neg2 * polygamma(1, alpha_neg1)
                + np.power(mu, 2) * alpha * np.power(1 + alphamu, -2)
                + alpha_neg2 * polygamma(1, y + alpha_neg1)
                + alpha_neg2 * y * np.power(mu + alpha_neg1, -2)
            )
        )
    else:
        ll_part = -2 * np.power(alpha, -3) * np.sum(
            digamma(alpha_neg1)
            + np.log(1 + alphamu)
            - alphamu / (1 + alphamu)
            - digamma(y + alpha_neg1)
            + y / (mu + alpha_neg1)
        ) + alpha_neg2 * np.sum(
            -1 * alpha_neg2 * polygamma(1, alpha_neg1)
            + np.power(mu, 2) * alpha * np.power(1 + alphamu, -2)
            + alpha_neg2 * polygamma(1, y + alpha_neg1)
            + alpha_neg2 * y * np.power(mu + alpha_neg1, -2)
        )

    # only the prior part is wrt log alpha
    if usePrior:
        prior_part = -1.0 / log_alpha_prior_sigmasq
    else:
        prior_part = 0.0

    # note: return (d2log_post/dalpha2 * alpha^2 + dlog_post/dalpha * alpha)
    #           =  (d2log_post/dalpha2 * alpha^2 + dlog_post/dlogalpha)
    # because we take derivatives wrt log alpha
    res = (
        (ll_part + cr_term) * np.power(alpha, 2)
        + dlog_posterior(
            log_alpha,
            y,
            mu,
            x,
            log_alpha_prior_mean,
            log_alpha_prior_sigmasq,
            False,
            weights,
            useWeights,
            weightThreshold,
            useCR,
        )
    ) + prior_part
    return res


def fitDisp(
    y,
    x,
    mu_hat,
    log_alpha,
    log_alpha_prior_mean,
    log_alpha_prior_sigmasq,
    min_log_alpha,
    kappa_0,
    tol,
    maxit,
    usePrior,
    weights,
    useWeights,
    weightThreshold,
    useCR,
):
    """
    Fit dispersions for negative binomial GLMs.

    This function estimates the dispersion parameter (alpha) for negative
    binomial generalized linear models. The fitting is performed on the log
    scale.

    Arguments
    ---------
    y : ndarray
        matrix of counts, shape (N,M)
    x : ndarray
        design matrix, shape (M,K)
    mu_hat : ndarray
        the expected mean values, shape (N,M)
    log_alpha : ndarray
        vector of initial guesses for log(alpha), shape N
    log_alpha_prior_mean : ndarray
        vector of fitted values for log(alpha), shape N
    log_alpha_prior_sigmasq : float
        the variance of the prior
    min_log_alpha : float
        the minimum value for log(alpha)
    kappa_0 : float
        parameter for initial proposal in the backtracking search.
        initial proposal = log(alpha) + kappa_0 * d(log-likelihood)/d(log(alpha))
    tol : float
        tolerance for convergence estimates
    maxit : int
        maximum number of iterations
    usePrior : bool
        whether to use a priori or just compute the MLE
    weights : ndarray
        observation weights, shape (N,M)
    useWeights : bool
        whether to use weights
    weightThreshold : float
        the threshold for subsetting the design matrix and GLM weights to
        calculate the Cox-Reid correction
    useCR : bool
        whether to use the Cox-Reid correction

    Returns
    -------
    log_alpha : ndarray
        the fitted dispersion parameters, on the log scale. Shape N.
    iter : ndarray
        the number of iterations for each parameter. Shape N.
    iter_accept : ndarray
        the number of accepted proposals for each parameter. Shape N.
    last_change : ndarray
        the last change of the fitted dispersion parameters. Shape N.
    initial_lp : ndarray
        the initial log posterior values. Shape N.
    initial_dlp : ndarray
        the initial derivatives (wrt. log(alpha)) of the log posterior. Shape N.
    last_lp : ndarray
        the last log posterior values. Shape N.
    last_dlp : ndarray
        the last derivatives (wrt. log(alpha)) of the log posterior. Shape N.
    last_d2lp : ndarray
        the last second derivatives (wrt. log(alpha)) of the log posterior. Shape N.
    """

    if isinstance(log_alpha, pd.Series):
        log_alpha = log_alpha.values
    if isinstance(log_alpha, (int, float)):
        log_alpha = np.repeat(float(log_alpha), y.shape[1])
    if isinstance(log_alpha_prior_mean, pd.Series):
        log_alpha_prior_mean = log_alpha_prior_mean.values
    if isinstance(log_alpha_prior_mean, (int, float)):
        log_alpha_prior_mean = np.repeat(float(log_alpha_prior_mean), y.shape[1])
    assert y.shape[1] == mu_hat.shape[1]
    assert y.shape[1] == log_alpha.shape[0]
    assert y.shape[1] == log_alpha_prior_mean.shape[0]

    y_n = y.shape[1]
    epsilon = 1.0e-4
    iter_ = np.zeros(y_n)
    iter_accept = np.zeros(y_n)

    # maximize the log likelihood over the variable a, the log of alpha, the
    # dispersion parameter.
    # in order to express the optimization in a typical manner, for calculating
    # theta(kappa) we multiply the log likelihood by -1 and seek a minimum
    # we use a line search based on the Armijo rule.
    # define a function theta(kappa) = f(a + kappa * d) where d is the search
    # direction
    # in this case the search direction is given by the first derivative of the
    # log likelihood
    a = log_alpha.copy()
    lp = log_posterior(
        a,
        y,
        mu_hat,
        x,
        log_alpha_prior_mean,
        log_alpha_prior_sigmasq,
        usePrior,
        weights,
        useWeights,
        weightThreshold,
        useCR,
    )
    dlp = dlog_posterior(
        a,
        y,
        mu_hat,
        x,
        log_alpha_prior_mean,
        log_alpha_prior_sigmasq,
        usePrior,
        weights,
        useWeights,
        weightThreshold,
        useCR,
    )

    kappa = np.repeat(kappa_0, y_n)
    initial_lp = lp.copy()
    initial_dlp = dlp.copy()
    change = np.repeat(-1.0, y_n)
    last_change = np.repeat(-1.0, y_n)

    idx = np.repeat(True, y_n)
    for t in range(maxit):
        # iter_ counts the number of steps taken out of maxit
        iter_[idx] += 1
        a_propose = a + kappa * dlp
        # note: lgamma is unstable for values around 1e17, where there is a
        # switch in lgamma.c
        # we limit log alpha from going lower than -30
        kappa = np.where(a_propose < -30.0, (-30.0 - a) / dlp, kappa)
        # we limit log alpha from going higher than 10
        kappa = np.where(a_propose > 10.0, (10.0 - a) / dlp, kappa)

        lpost = log_posterior(
            a[idx] + kappa[idx] * dlp[idx],
            y[:, idx],
            mu_hat[:, idx],
            x,
            log_alpha_prior_mean[idx],
            log_alpha_prior_sigmasq,
            usePrior,
            weights[:, idx],
            useWeights,
            weightThreshold,
            useCR,
        )
        theta_kappa = np.zeros(y_n)
        theta_kappa[idx] = -lpost
        theta_hat_kappa = -lp - kappa * epsilon * np.power(dlp, 2)

        # if this inequality is true, we have satisfied the Armijo rule and
        # accept the step size kappa, otherwise we halve kappa
        armijo_idx = idx & (theta_kappa <= theta_hat_kappa)
        # iter_accept counts the number of accepted proposals
        iter_accept[armijo_idx] += 1
        a[armijo_idx] = (a + kappa * dlp)[armijo_idx]
        lpnew = np.zeros(y_n)
        lpnew[idx] = lpost
        # look for change in log likelihood
        change[armijo_idx] = lpnew[armijo_idx] - lp[armijo_idx]
        lp = np.where(armijo_idx & (change < tol), lpnew, lp)
        idx[armijo_idx & (change < tol)] = False
        # if log(alpha) is going to -infinity, break the loop
        idx[armijo_idx & (a < min_log_alpha)] = False

        idx2 = idx & armijo_idx
        lp[idx2] = lpnew[idx2]
        dlp[idx2] = dlog_posterior(
            a[idx2],
            y[:, idx2],
            mu_hat[:, idx2],
            x,
            log_alpha_prior_mean[idx2],
            log_alpha_prior_sigmasq,
            usePrior,
            weights[:, idx2],
            useWeights,
            weightThreshold,
            useCR,
        )
        # instead of resetting kappa to kappa_0
        # multiply kappa by 1.1
        kappa[idx2] = np.minimum(kappa[idx2] * 1.1, kappa_0)
        # every 5 accepts, halve kappa
        # to prevent slow convergence due to overshooting
        kappa[idx2 & (iter_accept % 5 == 0)] /= 2.0

        kappa[~armijo_idx] /= 2.0

    last_lp = lp
    last_dlp = dlp
    last_d2lp = d2log_posterior(
        a,
        y,
        mu_hat,
        x,
        log_alpha_prior_mean,
        log_alpha_prior_sigmasq,
        usePrior,
        weights,
        useWeights,
        weightThreshold,
        useCR,
    )
    log_alpha = a
    # last change indicates the change for the final iteration
    last_change = change

    return {
        "log_alpha": log_alpha,
        "iter": iter_,
        "iter_accept": iter_accept,
        "last_change": last_change,
        "initial_lp": initial_lp,
        "initial_dlp": initial_dlp,
        "last_lp": last_lp,
        "last_dlp": last_dlp,
        "last_d2lp": last_d2lp,
    }


def fitDispWrapper(**kwargs):
    """
    Wrapper around :func:`fitDisp` to check for NaN in arguments

    See also
    --------
    fitDisp
    """
    for k, v in kwargs.items():
        if np.any(np.isnan(v)):
            raise ValueError(f"argument {k} of fitDisp contains a NaN value")
    return fitDisp(**kwargs)


def fitDispGrid(
    y,
    x,
    mu_hat,
    disp_grid,
    log_alpha_prior_mean,
    log_alpha_prior_sigmasq,
    usePrior,
    weights,
    useWeights,
    weightThreshold,
    useCR,
):
    """
    Fit dispersions by evaluating over a grid

    This function estimates the dispersion parameters (alpha) for negative
    binomial generalized linear models. The fitting is performed on the log
    scale.

    Arguments
    ---------
    y : ndarray
        matrix of counts, shape (N,M)
    x : ndarray
        design matrix, shape (M,K)
    mu_hat : ndarray
        the expected mean values, shape (N,M)
    disp_grid : ndarray
        the grid over which to estimate
    log_alpha_prior_mean : ndarray
        vector of fitted values for log(alpha), shape N
    log_alpha_prior_sigmasq : float
        the variance of the prior
    usePrior : bool
        whether to use a priori or just compute the MLE
    weights : ndarray
        observation weights, shape (N,M)
    useWeights : bool
        whether to use weights
    weightThreshold : float
        the threshold for subsetting the design matrix and GLM weights to
        calculate the Cox-Reid correction
    useCR : bool
        whether to use the Cox-Reid correction

    Returns
    -------
    ndarray
        the estimated dispersion parameters, on the log scale. Shape N.
    """
    y_n = y.shape[1]
    disp_grid_n = disp_grid.shape[0]
    delta = disp_grid[1] - disp_grid[0]

    logpostvec = log_posterior(
        disp_grid[:, None],
        y,
        mu_hat,
        x,
        log_alpha_prior_mean,
        log_alpha_prior_sigmasq,
        usePrior,
        weights,
        useWeights,
        weightThreshold,
        useCR,
    )
    idxmax = np.argmax(logpostvec, axis=0)
    assert idxmax.shape == (y_n,)
    a_hat = disp_grid[idxmax]
    disp_grid_fine = np.linspace(a_hat - delta, a_hat + delta, disp_grid_n)
    assert disp_grid_fine.shape == (disp_grid_n, y_n)
    logpostvec = log_posterior(
        disp_grid_fine,
        y,
        mu_hat,
        x,
        log_alpha_prior_mean,
        log_alpha_prior_sigmasq,
        usePrior,
        weights,
        useWeights,
        weightThreshold,
        useCR,
    )
    idxmax = np.argmax(logpostvec, axis=0)
    assert idxmax.shape == (y_n,)
    log_alpha = np.take_along_axis(disp_grid_fine, idxmax[None], axis=0).squeeze(0)
    assert log_alpha.shape == (y_n,)
    return log_alpha


def fitDispGridWrapper(**kwargs):
    """
    Wrapper around :func:`fitDispGrid`

    This wrapper checks for NaN in arguments, and automatically builds a grid
    on which :func:`fitDispGrid` will be called. Contrary to
    :func:`fitDispGrid`, it returns the estimated dispersion parameters on the
    natural scale.

    See also
    --------
    fitDispGrid

    Returns
    -------
    ndarray
        the estimated dispersion parameters, on the natural scale. Shape N.
    """
    for k, v in kwargs.items():
        if np.any(np.isnan(v)):
            raise ValueError(f"argument {k} of fitDispGrid contains a NaN value")

    minLogAlpha = np.log(1e-8)
    maxLogAlpha = np.log(np.maximum(10, kwargs["y"].shape[0]))
    dispGrid = np.linspace(minLogAlpha, maxLogAlpha, 20)
    kwargs["mu_hat"] = kwargs["mu"]
    del kwargs["mu"]
    kwargs["disp_grid"] = dispGrid
    logAlpha = fitDispGrid(**kwargs)
    return np.exp(logAlpha)


def fitBeta(
    y,
    x,
    nf,
    alpha_hat,
    contrast,
    beta_mat,
    lambda_,
    weights,
    useWeights,
    tol,
    maxit,
    useQR,
    minmu,
):
    """
    Fit beta coefficients for negative binomial GLMs

    This function estimates the coefficients (beta) for negative binomial
    generalized linear models. Fitting is performed on the log scale.

    Arguments
    ---------
    y : ndarray
        matrix of counts, shape (N,M)
    x : ndarray
        design matrix, shape (M,K)
    nf : ndarray
        matrix of normalization factors, shape (N,M)
    alpha_hat : ndarray
        vector of the dispersion estimates, shape N
    contrast : array-like
        vector for a possible contrast, shape K
    beta_mat : ndarray
        the initial estimates for the betas, shape (N,K)
    lambda_ : ndarray
        the ridge values, shape K
    weights : ndarray
        observation weights, shape (N,M)
    useWeights : bool
        whether to use weights
    tol : float
        tolerance for convergence estimates
    maxit : int
        maximum number of iterations
    useQR : bool
        whether to use QR decomposition

    Returns
    -------
    beta_mat : ndarray
        the fitted coefficients, on the log scale. Shape (N,K)
    beta_var_mat : ndarray
        the variance of the fitted coefficients. Shape (N,K)
    iter : ndarray
        the number of iterations for each row. Shape N
    hat_diagonals : ndarray
        TODO
    contrast_num : ndarray
        TODO
    contrast_denom : ndarray
        TODO
    deviance : ndarray
        TODO
    """
    y_m, y_n = y.shape
    x_p = x.shape[1]

    assert not isinstance(alpha_hat, pd.Series)
    assert alpha_hat.shape == (y_n,)
    assert beta_mat.shape == (y_n, x_p)
    assert y.shape[0] == x.shape[0]
    assert nf.shape == y.shape
    assert weights.shape == y.shape
    assert lambda_.ndim == 1
    assert lambda_.shape[0] == x.shape[1], f"{lambda_.shape}, {x.shape}"

    beta_var_mat = np.zeros(beta_mat.shape)
    contrast_num = np.zeros(beta_mat.shape[0])
    contrast_denom = np.zeros(beta_mat.shape[0])
    # bound the estimated count, as weights include 1/mu
    large = 30.0
    iter_ = np.zeros(y_n)
    ridge = np.diag(lambda_)

    mu_hat = np.maximum(nf * np.exp(x @ beta_mat.T), minmu)
    assert mu_hat.shape == nf.shape
    dev = np.zeros(y_n)
    dev_old = np.zeros(y_n)
    idx = np.repeat(True, y_n)

    for t in range(maxit):
        idx_n = np.sum(idx)
        iter_[idx] += 1

        mu_hat_idx = mu_hat[:, idx]
        if useWeights:
            w_vec = weights[:, idx] * mu_hat_idx / (1.0 + alpha_hat[idx] * mu_hat_idx)
        else:
            w_vec = mu_hat_idx / (1.0 + alpha_hat[idx] * mu_hat_idx)
        assert w_vec.shape == (y_m, idx_n)
        w_sqrt_vec = np.sqrt(w_vec)

        if useQR:
            # make an orthonormal design matrix including the ridge penalty
            # prepare matrices
            weighted_x_ridge = np.hstack(
                [
                    x * w_sqrt_vec.T[:, :, None],
                    np.broadcast_to(np.sqrt(ridge), (idx_n, x_p, x_p)),
                ]
            )
            assert weighted_x_ridge.shape == (idx_n, y_m + x_p, x_p)
            q, r = np.linalg.qr(weighted_x_ridge)
            assert q.shape[0] == r.shape[0] == idx_n
            big_w_diag = np.ones((idx_n, y_m + x_p))
            big_w_diag[:, :y_m] = w_vec.T
            z = np.log(mu_hat_idx / nf[:, idx]) + (y[:, idx] - mu_hat_idx) / mu_hat_idx
            assert z.shape == (y_m, idx_n)
            w_diag = w_vec.copy()
            z_sqrt_w = z * np.sqrt(w_diag)
            big_z_sqrt_w = np.zeros((idx_n, y_m + x_p))
            big_z_sqrt_w[:, :y_m] = z_sqrt_w.T
            # IRLS with Q matrix for X
            gamma_hat = np.swapaxes(q, -1, -2) @ big_z_sqrt_w[:, :, None]
            beta_hat = np.linalg.solve(r, gamma_hat.squeeze(-1))
        else:
            # use the standard design matrix and matrix inversion
            z = np.log(mu_hat_idx / nf[:, idx]) + (y[:, idx] - mu_hat_idx) / mu_hat_idx
            assert (x.T @ (x * w_vec.T[:, :, None]) + ridge).shape == (idx_n, x_p, x_p)
            assert ((z * w_vec).T @ x).shape == (idx_n, x_p)
            beta_hat = np.linalg.solve(
                x.T @ (x * w_vec.T[:, :, None]) + ridge, (z * w_vec).T @ x
            )

        assert beta_hat.shape == (idx_n, x_p), f"{beta_hat.shape} vs {(idx_n, x_p)}"
        # easier to update beta_mat here instead than at the end of the loop
        beta_mat[idx, :] = beta_hat
        subidx = ~(np.sum(np.abs(beta_hat) > large, axis=1) > 0)
        assert subidx.shape == (idx_n,)
        newidx = idx.copy()
        newidx[idx] = subidx
        iter_[idx & ~newidx] = maxit

        idx = newidx
        mu_hat[:, idx] = np.maximum(
            nf[:, idx] * np.exp(x @ beta_hat[subidx, :].T), minmu
        )
        dev[idx] = 0.0
        if useWeights:
            dev[idx] -= 2.0 * np.sum(
                weights[:, idx]
                * dnbinom_mu(y[:, idx], 1.0 / alpha_hat[idx], mu_hat[:, idx], True),
                axis=0,
            )
        else:
            dev[idx] -= 2.0 * np.sum(
                dnbinom_mu(y[:, idx], 1.0 / alpha_hat[idx], mu_hat[:, idx], True),
                axis=0,
            )

        conv_test = np.abs(dev - dev_old) / (np.abs(dev) + 0.1)
        nanidx = np.isnan(conv_test)
        iter_[idx & nanidx] = maxit
        idx &= ~nanidx
        if t > 0:
            idx &= ~(conv_test < tol)
        dev_old = dev.copy()

    # recalculate w so that this is identical if we start with beta_hat
    if useWeights:
        w_vec = weights * mu_hat / (1.0 + alpha_hat * mu_hat)
    else:
        w_vec = mu_hat / (1.0 + alpha_hat * mu_hat)
    w_sqrt_vec = np.sqrt(w_vec)

    xw = x * w_sqrt_vec.T[:, :, None]
    assert xw.shape == (y_n, y_m, x_p)
    xtwxr_inv = np.linalg.inv(x.T @ (x * w_vec.T[:, :, None]) + ridge)
    assert xtwxr_inv.shape == (y_n, x_p, x_p)

    hat_diagonals = np.zeros(y.shape)
    # this is equivalent to (for all j):
    #   hat_diagonals[:,j] = np.diag(xw[j] @ xtwxr_inv[j] @ xw[j].T)
    # but it avoids computing full matrix products just to retrieve the diags
    for k in range(x_p):
        for l in range(x_p):
            hat_diagonals[:, :] += (xw[:, :, k] * xw[:, :, l]).T * xtwxr_inv[:, k, l]

    # sigma is the covariance matrix for the betas
    sigma = xtwxr_inv @ x.T @ (x * w_vec.T[:, :, None]) @ xtwxr_inv
    assert sigma.shape == (y_n, x_p, x_p)
    contrast_num = contrast @ beta_mat.T
    contrast_denom = np.sqrt(contrast.T @ sigma @ contrast)
    beta_var_mat = np.diagonal(sigma, axis1=-2, axis2=-1)

    return {
        "beta_mat": beta_mat,
        "beta_var_mat": beta_var_mat,
        "iter": iter_,
        "hat_diagonals": hat_diagonals,
        "contrast_num": contrast_num,
        "contrast_denom": contrast_denom,
        "deviance": dev,
    }


def fitBetaWrapper(**kwargs):
    """
    Wrapper around :func:`fitBeta`

    This wrapper checks for NaN in arguments. It also sets a default contrast
    if none is provided.

    See also
    --------
    fitBeta
    """
    for k, v in kwargs.items():
        if np.any(np.isnan(v)):
            raise ValueError(f"argument {k} of fitBeta contains a NaN value")

    if "contrast" not in kwargs:
        kwargs["contrast"] = np.zeros(kwargs["x"].shape[1])
        kwargs["contrast"][0] = 1

    return fitBeta(**kwargs)



# distutils: language = c++
#-----------------------------------------------------------------------------
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import numpy as np
from scipy.linalg.cython_lapack cimport dormqr, dgeqrf, dtrtrs, dpotrf, dpotrs, dsytrf
from scipy.linalg.cython_blas cimport dgemv

cdef public object make_levenberg_result "make_levenberg_result"(
        ndarray[double, ndim=2] coefficients,
        ndarray[double, ndim=2] fitted_values,
        vector.vector[double] deviance,
        vector.vector[long] iter,
        vector.vector[char] failed):
    return (coefficients, fitted_values, deviance, iter, failed)

cdef public ndarray[double, ndim=1] vector2ndarray "vector2ndarray"(const vector.vector[double]& data):
    return np.asarray(data, order='F')

cdef char side = b'L'
cdef char trans_ormqr = b'T'
cdef char uplo = b'U'
cdef char trans_trtrs = b'N'
cdef char diag = b'N'
cdef int unity = 1

cdef public void build_QRdecomposition(QRdecomposition* self):
    # Setting up the workspace for dgeqrf
    cdef double tmpwork
    dgeqrf(&self.NR, &self.NC, self.Xcopy.data(), &self.NR, self.tau.data(), &tmpwork, &self.lwork_geqrf, &self.info)

    # Loading up the optimal WORK
    self.lwork_geqrf = <int>(tmpwork + 0.5)
    if self.lwork_geqrf < 1:
        self.lwork_geqrf = 1
    self.work_geqrf.resize(self.lwork_geqrf)

    # Repeating for dormqr
    dormqr(&side, &trans_ormqr, &self.NR, &unity, &self.NC, self.Xcopy.data(), &self.NR, self.tau.data(), self.effects.data(), &self.NR, &tmpwork, &self.lwork_ormqr, &self.info)

    self.lwork_ormqr = <int>(tmpwork + 0.5)
    if self.lwork_ormqr < 1:
        self.lwork_ormqr = 1
    self.work_ormqr.resize(self.lwork_ormqr)

cdef void decompose_QRdecomposition(QRdecomposition* self):
    dgeqrf(&self.NR, &self.NC, self.Xcopy.data(), &self.NR, self.tau.data(), self.work_geqrf.data(), &self.lwork_geqrf, &self.info)

cdef void solve_QRdecomposition(QRdecomposition* self):
    dormqr(&side, &trans_ormqr, &self.NR, &unity, &self.NC, self.Xcopy.data(), &self.NR, self.tau.data(), self.effects.data(), &self.NR, self.work_ormqr.data(), &self.lwork_ormqr, &self.info)
    if self.info:
        raise RuntimeError("Q**T multiplication failed")

    dtrtrs(&uplo, &trans_trtrs, &diag, &self.NC, &unity, self.Xcopy.data(), &self.NR, self.effects.data(), &self.NR, &self.info)
    if self.info:
        raise RuntimeError("failed to solve the triangular system")

cdef public void f77_dgemv "f77_dgemv"(const char *trans, const int *m, const int *n, const double *alpha, const double *a, const int *lda, const double *x, const int *incx, const double *beta, double *y, const int *incy) nogil:
    dgemv(<char*>trans, <int*>m, <int*>n, <double*>alpha, <double*>a, <int*>lda, <double*>x, <int*>incx, <double*>beta, y, <int*>incy)
cdef public void f77_dpotrf "f77_dpotrf"(const char *uplo, const int *n, double *a, const int *lda, int *info) nogil:
    dpotrf(<char*>uplo, <int*>n, a, <int*>lda, info)
cdef public void f77_dpotrs "f77_dpotrs"(const char *uplo, const int *n, const int *nrhs, const double *a, const int *lda, double *b, const int *ldb, int *info) nogil:
    dpotrs(<char*>uplo, <int*>n, <int*>nrhs, <double*>a, <int*>lda, b, <int*>ldb, info)
cdef public void f77_dsytrf "f77_dsytrf"(const char *uplo, const int *n, double *a, const int *lda, int *ipiv, double *work, const int *lwork, int *info) nogil:
    dsytrf(<char*>uplo, <int*>n, a, <int*>lda, ipiv, work, <int*>lwork, info)



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import logging
import numpy as np
import pandas as pd
from scipy.stats import trim_mean

from ..utils import Factor, pnorm, pt
from .fitNbinomGLMs import fitNbinomGLMs, fitGLMsWithPrior
from .misc import buildMatrixWithNACols, nOrMoreInCell, buildDataFrameWithNACols
from .weights import getAndCheckWeights


def nbinomWaldTest(
    obj,
    betaPrior=False,
    betaPriorVar=None,
    modelMatrix=None,
    modelMatrixType=None,
    betaTol=1e-8,
    maxit=100,
    useOptim=True,
    quiet=False,
    useT=False,
    df=None,
    useQR=True,
    minmu=0.5,
):
    """
    Wald test for the GLM coefficients

    This function tests for significance of coefficients in a Negative Binomial
    GLM, using previously calculated sizeFactors or normalizationFactors and
    dispersion estimates. See :func:`DESeq` for the GLM formula.

    The fitting proceeds as follows: standard maximum likelihood estimates for
    GLM coefficients (synonymous with "beta", "log2 fold change", "effect
    size") are calculated. Then, optionally, a zero-centered normal prior
    distribution (flag :code:`betaPrior`) is assumed for the coefficients other
    than the intercept.

    Note that this posterior log2 fold change estimation is now not the default
    setting for :func:`nbinomWaldTest`, as the standard workflow for
    coefficient shrinkage has moved to an additional function
    :func:`lfcShrink`.

    To calculate Wald test p-values, the coefficients are scaled by their
    standard errors and then compared to a standard normal distribution. The
    :meth:`.DESeqDataSet.results` method without any argument will
    automatically perform a contrast of the last level of the last variable in
    the design formula over the first level. The contrast argument of
    :meth:`.DESeqDataSet.results` can be used to generate other comparisons.

    The Wald test can be replaced with :func:`nbinomLRT` for an alternative
    test of significance.

    Notes
    -----
    The variance of the prior distribution for each non-intercept coefficient
    is calculated using the observed distribution of the maximum likelihood
    coefficients. The final coefficients are then maximum a posterior estimates
    using this prior (Tikhonov/ridge regularization).  See below for details on
    the prior variance and the methods section of the original DESeq2
    manuscript for more details. The use of a prior has little effect on genes
    with high counts and helps to moderate the large spread in coefficients for
    genes with low counts.

    The prior variance is calculated by matching the 0.05 upper quantile of the
    observed MLE coefficients to a zero-centered normal distribution. In a
    change of methods since the 2014 paper, the weighted upper quantile is
    calculated using the :func:`wtd_quantile` function from the Hmisc package
    (function has been copied into DESeq2 code to avoid extra dependencies).
    The weights are the inverse of the expected variance of log counts, so the
    inverse of :math:`1 / \overline{\mu} + \\alpha_{tr}` using the mean of
    normalized counts and the trended dispersion fit. The weighting ensures
    that noisy estimates of log fold changes from small count genes do not
    overly influence the calculation of the prior variance (see
    :func:`estimateBetaPriorVar`).  The final prior variance for a factor level
    is the average of the estimated prior variance over all contrasts of all
    levels of the factor.

    When a log2 fold change prior is used (:code:`betaPrior=True`), then
    :func:`nbinomWaldTest` will by default use expanded model matrices, as
    described in the :code:`modelMatrixType` argument, unless this arguments is
    used to override the default behavior.  This ensures that log2 fold changes
    will be independent of the choice of the reference level. In this case, the
    beta prior variance for each factor is calculated as the average of the
    mean squared maximum likelihood estimates for each level and every possible
    contrast.

    Arguments
    ---------
    obj : DESeqDataSet
        a DESeqDataSet object
    betaPrior : bool
        whether to use a zero-mean normal prior on the non-intercept
        coefficients
    betaPriorVar : ndarray, optional
        a vector with length equal to the number of model terms including the
        intercept, giving the variance of the prior on the sample beta on the
        log2 scale. If None, it is estimated from the data
    modelMatrix : ndarray, optional
        a design matrix. Typically left None and created by the function
    modelMatrixType : str
        either :code:`"standard"` or :code:`"expanded"`, which describe how the
        model matrix is formed.

        :code:`"standard"` means as created by :func:`patsy.dmatrix` using the
        design formula.

        :code:`"expanded"` includes an indicator variable for each level of
        factors in addition to an intercept. :code:`betaPrior` must be set to
        :code:`True` in order for expanded model matrices to be fit.
    betaTol : float
        control parameter defining convergence
    maxit : int
        the maximum number of iterations to allow for convergence of the
        coefficient vector
    useOptim : bool
        whether to use the native optimization function on rows that do not
        converged withing :code:`maxit` iterations
    quiet : bool
        whether to print messages at each step
    useT : bool
        whether to use a t-distribution as a null distribution, for
        significance testing of the Wald statistics.  If False, a standard
        normal null distribution is used. See next argument df for information
        about which t is used. If :code:`useT=True` then further calls to
        :meth:`.DESeqDataSet.results` will make use of
        :code:`obj.var["tDegreesFreedom"]` that is stored by
        :func:`nbinomWaldTest`.
    df : array-like
        the degrees of freedom for the t-distribution. It must be broadcastable
        to the number of columns of obj. If not specified, the degrees of
        freedom will be set by the number of samples minus the number of columns
        of the design matrix used for dispersion estimation. If weights are
        included in :code:`obj.layers`, then the sum of the weights is used in
        lieu of the number of samples.
    useQR : bool
        whether to use the QR decomposition of the design matrix while fitting
        the GLM
    minmu : float
        lower bound on the estimated count while fitting the GLM

    Returns
    -------
    DESeqDataSet
        the input DESeqDataSet with results columns accessible with the
        :meth:`.DESeqDataSet.results` method. The coefficients and standard
        errors are reported on a log2 scale.
    """

    if not quiet:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    if "dispersion" not in obj.var:
        raise ValueError(
            "testing requires dispersion estimates, first call estimateDispersions()"
        )

    if not obj.var.type.filter("results").empty:
        logging.info("found results columns, replacing these")
        obj = obj.removeResults()

    if obj.var["allZero"] is None:
        obj = obj.getBaseMeansAndVariances()

    # only continue on the columns with non-zero means
    objNZ = obj[:, ~obj.var["allZero"]]

    # model matrix not provided...
    if modelMatrix is None:
        modelAsFormula = True
        termsOrder = np.array([len(t.factors) for t in obj.design.design_info.terms])
        interactionPresent = (termsOrder > 1).any()

        # run some tests common to DESeq, nbinomWaldTest, nbinomLRT
        obj.designAndArgChecker(betaPrior)

        # what kind of model matrix to use
        if not isinstance(betaPrior, bool):
            raise ValueError("betaPrior must be a boolean")
        blindDesign = np.array_equal(termsOrder, [0])
        if blindDesign:
            betaPrior = False
        if modelMatrixType is None:
            if betaPrior:
                modelMatrixType = "expanded"
            else:
                modelMatrixType = "standard"
        if modelMatrixType == "expanded" and not betaPrior:
            raise ValueError("expanded model matrices require a beta prior")

        # store modelMatrixType so it can be accessed by estimateBetaPriorVar
        obj.modelMatrixType = modelMatrixType
        hasIntercept = 0 in termsOrder
        renameCols = hasIntercept
    else:  # modelMatrix is not None, user-supplied
        if betaPrior:
            if betaPriorVar is None:
                raise ValueError(
                    "user-supplied model matrix with betaPrior=True requires supplying betaPriorVar"
                )
        modelAsFormula = False
        obj.modelMatrixType = "user-supplied"
        renameCols = False

    if not betaPrior:
        # fit the negative binomial GLM without a prior
        # (in actuality a very wide prior with standard deviation 1e3 on log2 fold changes)
        fit = fitNbinomGLMs(
            objNZ,
            betaTol=betaTol,
            maxit=maxit,
            useOptim=useOptim,
            useQR=useQR,
            renameCols=renameCols,
            modelMatrix=modelMatrix,
            minmu=minmu,
        )
        H = fit["hat_diagonals"]
        mu = fit["mu"]
        modelMatrix = fit["modelMatrix"]
        # record the wide prior variance which was used in fitting
        betaPriorVar = np.full(modelMatrix.shape[1], 1e6)
    else:
        priorFitList = fitGLMsWithPrior(
            obj=obj,
            betaTol=betaTol,
            maxit=maxit,
            useOptim=useOptim,
            useQR=useQR,
            betaPriorVar=betaPriorVar,
            modelMatrix=modelMatrix,
            minmu=minmu,
        )
        fit = priorFitList["fit"]
        H = priorFitList["H"]
        mu = priorFitList["mu"]
        modelMatrix = priorFitList["modelMatrix"]
        betaPriorVar = priorFitList["betaPriorVar"]
        mleBetaMatrix = priorFitList["mleBetaMatrix"]

        # will add the MLE beta, so remove any which exist already
        # (possibly coming from estimateMLEForBetaPriorVar)
        MLEcols = obj.var.filter(regex="MLE_").columns
        for c in MLEcols:
            del obj.var[c]
            del obj.var.type[c]
            del obj.var.description[c]

    # store mu and H, the hat matrix diagonals
    objNZ.layers["mu"] = mu
    obj.layers["mu"] = buildMatrixWithNACols(mu, obj.var["allZero"])
    objNZ.layers["H"] = H
    obj.layers["H"] = buildMatrixWithNACols(H, obj.var["allZero"])

    # store the prior variance directly as an attribute of the DESeqDataSet
    # object, so it can be pulled later by the results functions
    # (necessary for setting max Cook's distance)
    obj.betaPrior = betaPrior
    obj.betaPriorVar = betaPriorVar
    obj.modelMatrix = modelMatrix
    obj.test = "Wald"

    # compute Cook's distance
    if modelAsFormula:
        dispModelMatrix = obj.design
    else:
        dispModelMatrix = modelMatrix
    obj.dispModelMatrix = dispModelMatrix
    cooks = calculateCooksDistance(objNZ, H, dispModelMatrix)

    # record maximum Cook's
    maxCooks = recordMaxCooks(obj.design, obj.obs, dispModelMatrix, cooks, objNZ.n_vars)

    # store Cook's distance for each sample
    obj.layers["cooks"] = buildMatrixWithNACols(cooks, obj.var["allZero"])

    # add betas, standard errors and Wald p-values to the object
    modelMatrixNames = modelMatrix.design_info.column_names
    betaMatrix = fit["betaMatrix"]
    assert isinstance(betaMatrix, pd.DataFrame)
    assert betaMatrix.shape == (
        objNZ.n_vars,
        len(modelMatrixNames),
    ), f"betaMatrix shape {betaMatrix.shape} is wrong, should be {(objNZ.n_vars, len(modelMatrixNames))}"
    betaMatrix.index = objNZ.var_names
    assert np.array_equal(betaMatrix.columns, modelMatrixNames)
    betaSE = fit["betaSE"]
    assert isinstance(betaSE, pd.DataFrame)
    assert (
        betaSE.shape == betaMatrix.shape
    ), f"betaSE and betaMatrix shapes disagree: {betaSE.shape} and {betaMatrix.shape}"
    betaSE.index = objNZ.var_names
    betaSE.columns = [f"SE_{n}" for n in modelMatrixNames]
    WaldStatistic = betaMatrix / betaSE.values
    WaldStatistic.columns = [f"WaldStatistic_{n}" for n in modelMatrixNames]

    #################################
    ## t distribution for p-values ##
    #################################

    if useT:
        # if the `df` was provided to nbinomWaldTest...
        if df is not None:
            df = np.asarray(df)
            if len(df) != obj.n_vars:
                raise ValueError(
                    "df should have the same length as the number of vars in obj"
                )
            if len(df.shape) == 0:
                df = np.repeat(df, objNZ.n_vars)
            else:
                # the WaldStatistic vector is along nonzero cols of obj
                df = df[~obj.var["allZero"]]
        else:
            # df is missing, so compute it from the number of samples (wrt weights)
            # and the number of coefficients
            if "weights" in obj.layers:
                # this checks that weights are OK and normalizes to have colMax == 1
                # (although this has already happened earlier in estDispGeneEst and estDispMAP...)
                (_, weights, _) = getAndCheckWeights(objNZ, dispModelMatrix)
                num_samps = np.sum(weights, axis=0)
            else:
                num_samps = np.repeat(obj.n_obs, objNZ.n_vars)

            df = num_samps - dispModelMatrix.shape[1]

        df = np.where(df > 0, df, np.nan)
        if df.shape[0] != WaldStatistic.shape[0]:
            raise RuntimeError(
                f"df and WaldStatistic have incompatible shapes: {df.shape} vs {WaldStatistic.shape}"
            )
        # use a t distribution to calculate the p-value
        WaldPvalue = 2 * pt(np.abs(WaldStatistic), df=df[:, None], lower_tail=False)
    else:
        WaldPvalue = 2 * pnorm(np.abs(WaldStatistic), lower_tail=False)

    WaldPvalue = pd.DataFrame(
        WaldPvalue,
        index=objNZ.var_names,
        columns=[f"WaldPvalue_{n}" for n in modelMatrixNames],
    )

    betaConv = fit["betaConv"]

    if np.any(~betaConv):
        logging.info(
            f"{np.sum(~betaConv)} cols did not converge in beta, labelled in obj.var['betaConv']. Use larger maxit argument with nbinomWaldTest"
        )

    resultsList = [betaMatrix, betaSE, WaldStatistic, WaldPvalue]
    if betaPrior:
        resultsList.append(mleBetaMatrix)
    resultsDF = pd.concat(resultsList, axis=1)

    resultsDF["betaConv"] = betaConv
    resultsDF["betaIter"] = fit["betaIter"]
    resultsDF["deviance"] = -2 * fit["logLike"]
    resultsDF["maxCooks"] = maxCooks

    # if useT need to add the t degrees of freedom to the end of resultsList
    if useT:
        resultsDF["tDegreesFreedom"] = df

    WaldResults = buildDataFrameWithNACols(resultsDF, obj.var["allZero"])
    WaldResults.index = obj.var_names
    assert np.sum(WaldResults.columns.isin(obj.var.columns)) == 0
    new_var = pd.concat([obj.var, WaldResults], axis=1)
    for c in obj.var.columns:
        new_var.type[c] = obj.var.type[c]
        new_var.description[c] = obj.var.description[c]
    obj.var = new_var

    for c in WaldResults.columns:
        obj.var.type[c] = "results"

    obj.var.description["betaConv"] = "convergence of betas"
    obj.var.description["betaIter"] = "iterations for betas"
    obj.var.description["deviance"] = "deviance for the fitted model"
    obj.var.description["maxCooks"] = "maximum Cook's distance for column"

    lfcType = "MAP" if obj.betaPrior else "MLE"
    for c, n in zip(betaMatrix.columns, modelMatrixNames):
        obj.var.description[c] = f"log2 fold change ({lfcType}): {n}"
    for c, n in zip(betaSE.columns, modelMatrixNames):
        obj.var.description[c] = f"standard error: {n}"
    for c, n in zip(WaldStatistic.columns, modelMatrixNames):
        obj.var.description[c] = f"Wald statistic: {n}"

    for c, n in zip(WaldPvalue.columns, modelMatrixNames):
        obj.var.description[c] = f"Wald test p-value: {n}"
    if betaPrior:
        for c, n in zip(mleBetaMatrix.columns, modelMatrixNames):
            obj.var.description[c] = c.replace("_", " ")
    if useT:
        obj.var.description["tDegreesFreedom"] = "t degrees of freedom for Wald test"

    return obj


def calculateCooksDistance(obj, H, modelMatrix):
    """
    Compute Cook's distance

    Arguments
    ---------
    obj : DESeqDataSet
        the object on which to compute the Cook's distance
    H : ndarray
    modelMatrix : ndarray
        the design matrix
    """
    p = modelMatrix.shape[1]
    dispersions = robustMethodOfMomentsDisp(obj, modelMatrix)
    V = obj.layers["mu"] + dispersions * obj.layers["mu"] ** 2
    PearsonResSq = (obj.counts() - obj.layers["mu"]) ** 2 / V
    return PearsonResSq / p * H / (1 - H) ** 2


# TODO make it a method of DESeqDataSet
def robustMethodOfMomentsDisp(obj, modelMatrix):
    """
    A robust method of moments dispersion

    This function estimates the dispersion excluding individual outlier counts,
    which would raise the variance estimate.

    Arguments
    ---------
    obj : DESeqDataSet
        a DESeqDataSet object
    modelMatrix : matrix
        a design matrix

    Returns
    -------
    vector
        estimates of moments dispersion
    """
    cnts = obj.counts(normalized=True)
    # if there are 3 or more replicates in any cell
    threeOrMore = nOrMoreInCell(modelMatrix, n=3)
    if np.any(threeOrMore):
        cells = Factor([tuple(modelMatrix[i]) for i in range(modelMatrix.shape[0])])
        levelsThreeOrMore = cells.categories[cells.value_counts() >= 3]
        idx = cells.isin(levelsThreeOrMore)
        cntsSub = cnts[idx, :]
        cellsSub = Factor(cells[idx])
        v = trimmedCellVariance(cntsSub, cellsSub)
    else:
        v = trimmedVariance(cnts)

    m = np.mean(cnts, axis=0)
    alpha = (v - m) / m**2
    # cannot use the typical minDisp = 1e-8 here or else all counts in the same
    # group as the outlier count will get an extrem Cook's distance
    minDisp = 0.04
    alpha = np.maximum(alpha, minDisp)
    return alpha


def trimmedCellVariance(cnts, cells):
    """
    TODO
    """
    # how much to trim at different n
    trimratio = [1 / 3, 1 / 4, 1 / 8]

    # returns an index for the vector above for three sample size bins
    def trimfn(n):
        if n <= 0:
            return np.nan
        elif n <= 3.5:
            return 0
        elif n <= 23.5:
            return 1
        else:
            return 2

    cellMeans = np.vstack(
        [
            trim_mean(cnts[cells == lvl, :], trimratio[trimfn(n)], axis=0)
            for lvl, n in cells.value_counts().items()
        ]
    )
    assert cellMeans.shape[1] == cnts.shape[1]
    assert cellMeans.shape[0] == len(cells.value_counts())

    qmat = cellMeans[cells.codes, :]
    sqerror = (cnts - qmat) ** 2
    varEst = np.vstack(
        [
            [2.04, 1.86, 1.51][trimfn(n)]
            * trim_mean(sqerror[cells == lvl, :], trimratio[trimfn(n)], axis=0)
            for lvl, n in cells.value_counts().items()
        ]
    )
    assert varEst.shape[1] == sqerror.shape[1]
    assert varEst.shape[0] == len(cells.value_counts())

    # take the max of variance estimates from cells
    # as one condition might have highly variable counts
    return np.max(varEst, axis=0)


def trimmedVariance(x):
    rm = trim_mean(x, 1 / 8, axis=0)
    sqerror = (x - rm) ** 2
    # scale due to trimming of large squares
    return 1.51 * trim_mean(sqerror, 1 / 8, axis=0)


def recordMaxCooks(design, clinicalData, modelMatrix, cooks, numCol):
    """this function breaks out the logic for calculating the max Cook's distance:
    the samples over which max Cook's distance is calculated:

    Cook's distance is considered for those samples with 3 or more replicates per cell

    if m == p or there are no samples over which to calculate max Cook's, return NA
    """
    samplesForCooks = nOrMoreInCell(modelMatrix, n=3)
    m, p = modelMatrix.shape
    if m > p and np.any(samplesForCooks):
        return np.max(cooks[samplesForCooks, :], axis=0)
    else:
        return np.repeat(np.nan, numCol)



# -----------------------------------------------------------------------------
# Copyright (C) 2022-2023 M. Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------


import numpy as np
from pandas import Categorical


class Factor(Categorical):
    """
    A class to represent a factor, as in R.

    It is essentially a :obj:`pandas.Categorical` object, with additional
    methods to mimic R API.
    """

    def __init__(self, arr):
        """
        Constructs a Factor instance from an array

        Arguments
        ---------
        arr : array_like
            a list of factors
        """

        super().__init__(arr)

    def droplevels(self):
        """
        drop unused levels
        """

        return Factor(self.__array__())

    def nlevels(self):
        """
        the number of levels
        """

        return len(self.categories)


def asfactor(g):
    if type(g) is Factor:
        return g
    else:
        return Factor(g)


def gl(n, k):
    arr = []
    for i in range(1, n + 1):
        arr.extend([i for j in range(k)])
    return Factor(arr)
